# Copyright 2023 DDPO-pytorch authors (Kevin Black), metric-space, The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
from collections import defaultdict
from concurrent import futures
from typing import Any, Callable, Optional, Tuple
from warnings import warn

import PIL.Image
import torch
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import ProjectConfiguration, set_seed
from huggingface_hub import whoami

from .base_trainer import BaseTrainer
from trl import DDPOConfig,DDPOPipelineOutput
from .per_prompt import PerPromptStatTracker
from .better_ddpo_pipeline import BetterDefaultDDPOStableDiffusionPipeline
import re
import wandb
import PIL
from torchvision.transforms.functional import to_pil_image
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion import rescale_noise_cfg,StableDiffusionPipeline
from trl.models.modeling_sd_base import scheduler_step
from typing import Any, Callable, Dict, List, Optional, Union 
from PIL import Image


logger = get_logger(__name__)


MODEL_CARD_TEMPLATE = """---
license: apache-2.0
tags:
- trl
- ddpo
- diffusers
- reinforcement-learning
- text-to-image
- stable-diffusion
---

# {model_name}

This is a diffusion model that has been fine-tuned with reinforcement learning to
 guide the model outputs according to a value, function, or human feedback. The model can be used for image generation conditioned with text.

"""


@torch.no_grad()
def pipeline_step_ip_adapter(    self:StableDiffusionPipeline,
    prompt: Optional[Union[str, List[str]]] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
    num_inference_steps: int = 50,
    guidance_scale: float = 7.5,
    negative_prompt: Optional[Union[str, List[str]]] = None,
    num_images_per_prompt: Optional[int] = 1,
    eta: float = 0.0,
    generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
    latents: Optional[torch.FloatTensor] = None,
    prompt_embeds: Optional[torch.FloatTensor] = None,
    negative_prompt_embeds: Optional[torch.FloatTensor] = None,
    output_type: Optional[str] = "pil",
    return_dict: bool = True,
    callback: Optional[Callable[[int, int, torch.FloatTensor], None]] = None,
    callback_steps: int = 1,
    cross_attention_kwargs: Optional[Dict[str, Any]] = None,
    guidance_rescale: float = 0.0,
    ip_adapter_image:Optional[Image.Image]=None
)->DDPOPipelineOutput:
        r"""
    Function invoked when calling the pipeline for generation.  Args: prompt (`str` or `List[str]`, *optional*): The prompt or prompts to guide the image generation. If not defined, one has to pass `prompt_embeds`.  instead.  height (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor): The height in pixels of the generated image.
        width (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor):
            The width in pixels of the generated image.
        num_inference_steps (`int`, *optional*, defaults to 50):
            The number of denoising steps. More denoising steps usually lead to a higher quality image at the
            expense of slower inference.
        guidance_scale (`float`, *optional*, defaults to 7.5):
            Guidance scale as defined in [Classifier-Free Diffusion Guidance](https://arxiv.org/abs/2207.12598).
            `guidance_scale` is defined as `w` of equation 2. of [Imagen
            Paper](https://arxiv.org/pdf/2205.11487.pdf). Guidance scale is enabled by setting `guidance_scale >
            1`. Higher guidance scale encourages to generate images that are closely linked to the text `prompt`,
            usually at the expense of lower image quality.
        negative_prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts not to guide the image generation. If not defined, one has to pass
            `negative_prompt_embeds` instead. Ignored when not using guidance (i.e., ignored if `guidance_scale` is
            less than `1`).
        num_images_per_prompt (`int`, *optional*, defaults to 1):
            The number of images to generate per prompt.
        eta (`float`, *optional*, defaults to 0.0):
            Corresponds to parameter eta (η) in the DDIM paper: https://arxiv.org/abs/2010.02502. Only applies to
            [`schedulers.DDIMScheduler`], will be ignored for others.
        generator (`torch.Generator` or `List[torch.Generator]`, *optional*):
            One or a list of [torch generator(s)](https://pytorch.org/docs/stable/generated/torch.Generator.html)
            to make generation deterministic.
        latents (`torch.FloatTensor`, *optional*):
            Pre-generated noisy latents, sampled from a Gaussian distribution, to be used as inputs for image
            generation. Can be used to tweak the same generation with different prompts. If not provided, a latents
            tensor will ge generated by sampling using the supplied random `generator`.
        prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting. If not
            provided, text embeddings will be generated from `prompt` input argument.
        negative_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, negative_prompt_embeds will be generated from `negative_prompt` input
            argument.
        output_type (`str`, *optional*, defaults to `"pil"`):
            The output format of the generate image. Choose between
            [PIL](https://pillow.readthedocs.io/en/stable/): `PIL.Image.Image` or `np.array`.
        return_dict (`bool`, *optional*, defaults to `True`):
            Whether or not to return a [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`] instead of a
            plain tuple.
        callback (`Callable`, *optional*):
            A function that will be called every `callback_steps` steps during inference. The function will be
            called with the following arguments: `callback(step: int, timestep: int, latents: torch.FloatTensor)`.
        callback_steps (`int`, *optional*, defaults to 1):
            The frequency at which the `callback` function will be called. If not specified, the callback will be
            called at every step.
        cross_attention_kwargs (`dict`, *optional*):
            A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
            `self.processor` in
            [diffusers.cross_attention](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/cross_attention.py).
        guidance_rescale (`float`, *optional*, defaults to 0.7):
            Guidance rescale factor proposed by [Common Diffusion Noise Schedules and Sample Steps are
            Flawed](https://arxiv.org/pdf/2305.08891.pdf) `guidance_scale` is defined as `φ` in equation 16. of
            [Common Diffusion Noise Schedules and Sample Steps are Flawed](https://arxiv.org/pdf/2305.08891.pdf).
            Guidance rescale factor should fix overexposure when using zero terminal SNR.

    Examples:

    Returns:
        `DDPOPipelineOutput`: The generated image, the predicted latents used to generate the image and the associated log probabilities
    """
        # 0. Default height and width to unet
        height = height or self.unet.config.sample_size * self.vae_scale_factor
        width = width or self.unet.config.sample_size * self.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(
            prompt,
            height,
            width,
            callback_steps,
            negative_prompt,
            prompt_embeds,
            negative_prompt_embeds,
        )

        # 2. Define call parameters
        if prompt is not None and isinstance(prompt, str):
            batch_size = 1
        elif prompt is not None and isinstance(prompt, list):
            batch_size = len(prompt)
        else:
            batch_size = prompt_embeds.shape[0]

        device = self._execution_device
        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        # 3. Encode input prompt
        text_encoder_lora_scale = cross_attention_kwargs.get("scale", None) if cross_attention_kwargs is not None else None
        prompt_embeds = self._encode_prompt(
            prompt,
            device,
            num_images_per_prompt,
            do_classifier_free_guidance,
            negative_prompt,
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            lora_scale=text_encoder_lora_scale,
        )

        added_cond_kwargs ={}
        if ip_adapter_image:
            image_embeds = self.prepare_ip_adapter_image_embeds(
                ip_adapter_image,
                None,
                device,
                batch_size * num_images_per_prompt,
                self.do_classifier_free_guidance,
            )
            added_cond_kwargs = {"image_embeds": image_embeds}

        # 4. Prepare timesteps
        self.scheduler.set_timesteps(num_inference_steps, device=device)
        timesteps = self.scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.config.in_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            prompt_embeds.dtype,
            device,
            generator,
            latents,
        )

        # 6. Denoising loop
        num_warmup_steps = len(timesteps) - num_inference_steps * self.scheduler.order
        all_latents = [latents]
        all_log_probs = []
        with self.progress_bar(total=num_inference_steps) as progress_bar:
            for i, t in enumerate(timesteps):
                # expand the latents if we are doing classifier free guidance
                latent_model_input = torch.cat([latents] * 2) if do_classifier_free_guidance else latents
                latent_model_input = self.scheduler.scale_model_input(latent_model_input, t)

                # predict the noise residual
                noise_pred = self.unet(
                    latent_model_input,
                    t,
                    encoder_hidden_states=prompt_embeds,
                    cross_attention_kwargs=cross_attention_kwargs,
                    return_dict=False,
                    added_cond_kwargs=added_cond_kwargs
                )[0]

                # perform guidance
                if do_classifier_free_guidance:
                    noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                    noise_pred = noise_pred_uncond + guidance_scale * (noise_pred_text - noise_pred_uncond)

                if do_classifier_free_guidance and guidance_rescale > 0.0:
                    # Based on 3.4. in https://arxiv.org/pdf/2305.08891.pdf
                    noise_pred = rescale_noise_cfg(noise_pred, noise_pred_text, guidance_rescale=guidance_rescale)

                # compute the previous noisy sample x_t -> x_t-1
                scheduler_output = scheduler_step(self.scheduler, noise_pred, t, latents, eta)
                latents = scheduler_output.latents
                log_prob = scheduler_output.log_probs

                all_latents.append(latents)
                all_log_probs.append(log_prob)

                # call the callback, if provided
                if i == len(timesteps) - 1 or ((i + 1) > num_warmup_steps and (i + 1) % self.scheduler.order == 0):
                    progress_bar.update()
                    if callback is not None and i % callback_steps == 0:
                        callback(i, t, latents)

        if not output_type == "latent":
            image = self.vae.decode(latents / self.vae.config.scaling_factor, return_dict=False)[0]
            image, has_nsfw_concept = self.run_safety_checker(image, device, prompt_embeds.dtype)
        else:
            image = latents
            has_nsfw_concept = None

        if has_nsfw_concept is None:
            do_denormalize = [True] * image.shape[0]
        else:
            do_denormalize = [not has_nsfw for has_nsfw in has_nsfw_concept]

        image = self.image_processor.postprocess(image, output_type=output_type, do_denormalize=do_denormalize)

        # Offload last model to CPU
        if hasattr(self, "final_offload_hook") and self.final_offload_hook is not None:
            self.final_offload_hook.offload()

        return DDPOPipelineOutput(image, all_latents, all_log_probs)

def get_image_sample_hook(image_dir):
    os.makedirs(image_dir, exist_ok=True)
    def _fn(prompt_image_data, global_step, tracker):
        for row in prompt_image_data:
            images=row[0]
            prompts=row[1]
            for img,pmpt in zip(images, prompts):
                pmpt=pmpt.replace(" ", "_")
                pmpt=re.sub(r'\W+', '', pmpt)
                pmpt=pmpt[:45]
                try:
                    tracker.log({f"{pmpt}":wandb.Image(img)},tracker.tracker.step)
                except:
                    path=image_dir+"/"+pmpt+".png"
                    print("saving at ",path)
                    img.save(path)
                    try:
                        tracker.log({f"{pmpt}":wandb.Image(path)},tracker.tracker.step)
                    except:
                        pass
    return _fn

class BetterDDPOTrainer(BaseTrainer):
    """
    The DDPOTrainer uses Deep Diffusion Policy Optimization to optimise diffusion models.
    Note, this trainer is heavily inspired by the work here: https://github.com/kvablack/ddpo-pytorch
    As of now only Stable Diffusion based pipelines are supported

    Attributes:
        **config** (`DDPOConfig`) -- Configuration object for DDPOTrainer. Check the documentation of `PPOConfig` for more
         details.
        **reward_function** (Callable[[torch.Tensor, Tuple[str], Tuple[Any]], torch.Tensor]) -- Reward function to be used
        **prompt_function** (Callable[[], Tuple[str, Any]]) -- Function to generate prompts to guide model
        **sd_pipeline** (`DDPOStableDiffusionPipeline`) -- Stable Diffusion pipeline to be used for training.
        **image_samples_hook** (Optional[Callable[[Any, Any, Any], Any]]) -- Hook to be called to log images
    """

    _tag_names = ["trl", "ddpo"]

    def __init__(
        self,
        config: DDPOConfig,
        reward_function: Callable[[torch.Tensor, Tuple[str], int, Tuple[Any]], torch.Tensor],
        prompt_function: Callable[[], Tuple[str, Any]],
        sd_pipeline: BetterDefaultDDPOStableDiffusionPipeline,
        image_samples_hook: Optional[Callable[[Any, Any, Any], Any]] = None,
        subject_key: Optional[str]="",
        image_size:Optional[int]=512,
        use_ip_adapter:Optional[bool]=False,
        ip_adapter_scale:Optional[dict]=None,
        ip_adapter_weight_name:Optional[str]="ip-adapter_sd15.bin",
        ip_adapter_src_image: Optional[PIL.Image.Image]=None
    ):
        if image_samples_hook is None:
            warn("No image_samples_hook provided; no images will be logged")

        self.prompt_fn = prompt_function
        self.reward_fn = reward_function
        self.config = config
        self.image_samples_callback = image_samples_hook
        self.subject_key=subject_key
        self.image_size=image_size
        self.use_ip_adapter=use_ip_adapter
        self.ip_adapter_src_image=ip_adapter_src_image

        accelerator_project_config = ProjectConfiguration(**self.config.project_kwargs)

        if self.config.resume_from:
            self.config.resume_from = os.path.normpath(os.path.expanduser(self.config.resume_from))
            if "checkpoint_" not in os.path.basename(self.config.resume_from):
                # get the most recent checkpoint in this directory
                checkpoints = list(
                    filter(
                        lambda x: "checkpoint_" in x,
                        os.listdir(self.config.resume_from),
                    )
                )
                if len(checkpoints) == 0:
                    raise ValueError(f"No checkpoints found in {self.config.resume_from}")
                checkpoint_numbers = sorted([int(x.split("_")[-1]) for x in checkpoints])
                self.config.resume_from = os.path.join(
                    self.config.resume_from,
                    f"checkpoint_{checkpoint_numbers[-1]}",
                )

                accelerator_project_config.iteration = checkpoint_numbers[-1] + 1

        # number of timesteps within each trajectory to train on
        self.num_train_timesteps = int(self.config.sample_num_steps * self.config.train_timestep_fraction)

        self.accelerator = Accelerator(
            log_with=self.config.log_with,
            mixed_precision=self.config.mixed_precision,
            project_config=accelerator_project_config,
            # we always accumulate gradients across timesteps; we want config.train.gradient_accumulation_steps to be the
            # number of *samples* we accumulate across, so we need to multiply by the number of training timesteps to get
            # the total number of optimizer steps to accumulate across.
            gradient_accumulation_steps=self.config.train_gradient_accumulation_steps * self.num_train_timesteps,
            **self.config.accelerator_kwargs,
        )

        is_okay, message = self._config_check()
        if not is_okay:
            raise ValueError(message)

        is_using_tensorboard = config.log_with is not None and config.log_with == "tensorboard"

        if self.accelerator.is_main_process:
            self.accelerator.init_trackers(
                self.config.tracker_project_name,
                config=dict(ddpo_trainer_config=config.to_dict()) if not is_using_tensorboard else config.to_dict(),
                init_kwargs=self.config.tracker_kwargs,
            )

        logger.info(f"\n{config}")

        set_seed(self.config.seed, device_specific=True)

        self.sd_pipeline = sd_pipeline

        self.sd_pipeline.set_progress_bar_config(
            position=1,
            disable=not self.accelerator.is_local_main_process,
            leave=False,
            desc="Timestep",
            dynamic_ncols=True,
        )

        # For mixed precision training we cast all non-trainable weights (vae, non-lora text_encoder and non-lora unet) to half-precision
        # as these weights are only used for inference, keeping weights in full precision is not required.
        if self.accelerator.mixed_precision == "fp16":
            inference_dtype = torch.float16
        elif self.accelerator.mixed_precision == "bf16":
            inference_dtype = torch.bfloat16
        else:
            inference_dtype = torch.float32

        if use_ip_adapter:
            self.sd_pipeline.sd_pipeline.load_ip_adapter("h94/IP-Adapter", subfolder="models", weight_name=ip_adapter_weight_name)
            if ip_adapter_scale is not None:
                self.sd_pipeline.sd_pipeline.set_ip_adapter_scale(ip_adapter_scale)
            self.sd_pipeline.sd_pipeline.image_encoder.to(self.accelerator.device, dtype=inference_dtype)
            #self.sd_pipeline.sd_pipeline.image_encoder

        self.sd_pipeline.vae.to(self.accelerator.device, dtype=inference_dtype)
        self.sd_pipeline.text_encoder.to(self.accelerator.device, dtype=inference_dtype)
        self.sd_pipeline.unet.to(self.accelerator.device, dtype=inference_dtype)

        trainable_layers = self.sd_pipeline.get_trainable_layers()

        self.accelerator.register_save_state_pre_hook(self._save_model_hook)
        self.accelerator.register_load_state_pre_hook(self._load_model_hook)

        # Enable TF32 for faster training on Ampere GPUs,
        # cf https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices
        if self.config.allow_tf32:
            torch.backends.cuda.matmul.allow_tf32 = True

        self.optimizer = self._setup_optimizer(
            trainable_layers.parameters() if not isinstance(trainable_layers, list) else trainable_layers
        )

        self.neg_prompt_embed = self.sd_pipeline.text_encoder(
            self.sd_pipeline.tokenizer(
                [""] if self.config.negative_prompts is None else self.config.negative_prompts,
                return_tensors="pt",
                padding="max_length",
                truncation=True,
                max_length=self.sd_pipeline.tokenizer.model_max_length,
            ).input_ids.to(self.accelerator.device)
        )[0]

        if config.per_prompt_stat_tracking:
            self.stat_tracker = PerPromptStatTracker(
                config.per_prompt_stat_tracking_buffer_size,
                config.per_prompt_stat_tracking_min_count,
            )

        # NOTE: for some reason, autocast is necessary for non-lora training but for lora training it isn't necessary and it uses
        # more memory
        self.autocast = self.sd_pipeline.autocast or self.accelerator.autocast

        self.trainable_layers, self.optimizer = self.accelerator.prepare(trainable_layers, self.optimizer)

        if self.config.async_reward_computation:
            self.executor = futures.ThreadPoolExecutor(max_workers=config.max_workers)

        if config.resume_from:
            logger.info(f"Resuming from {config.resume_from}")
            self.accelerator.load_state(config.resume_from)
            self.first_epoch = int(config.resume_from.split("_")[-1]) + 1
        else:
            self.first_epoch = 0

    def compute_rewards(self, prompt_image_pairs, epoch:int, is_async=False):
        if not is_async:
            rewards = []
            for images, prompts, prompt_metadata in prompt_image_pairs:
                reward, reward_metadata = self.reward_fn(images, prompts, epoch,prompt_metadata)
                rewards.append(
                    (
                        torch.as_tensor(reward, device=self.accelerator.device,dtype=torch.float32),
                        reward_metadata,
                    )
                )
        else:
            rewards = self.executor.map(lambda x: self.reward_fn(
                x[0],x[1],epoch,x[2]
            ), prompt_image_pairs)
            rewards = [
                (torch.as_tensor(reward.result(), device=self.accelerator.device,dtype=torch.float32), reward_metadata.result())
                for reward, reward_metadata in rewards
            ]

        return zip(*rewards)

    def step(self, epoch: int, global_step: int,retain_graph:bool,normalize_rewards:bool):
        """
        Perform a single step of training.

        Args:
            epoch (int): The current epoch.
            global_step (int): The current global step.

        Side Effects:
            - Model weights are updated
            - Logs the statistics to the accelerator trackers.
            - If `self.image_samples_callback` is not None, it will be called with the prompt_image_pairs, global_step, and the accelerator tracker.

        Returns:
            global_step (int): The updated global step.

        """
        samples, prompt_image_data = self._generate_samples(
            iterations=self.config.sample_num_batches_per_epoch,
            batch_size=self.config.sample_batch_size,
        )

        # collate samples into dict where each entry has shape (num_batches_per_epoch * sample.batch_size, ...)
        samples = {k: torch.cat([s[k] for s in samples]) for k in samples[0].keys()}
        rewards, rewards_metadata = self.compute_rewards(
            prompt_image_data, epoch,is_async=self.config.async_reward_computation
        )

        for i, image_data in enumerate(prompt_image_data):
            image_data.extend([rewards[i], rewards_metadata[i]])

        if self.image_samples_callback is not None:
            self.image_samples_callback(prompt_image_data, global_step, self.accelerator.trackers[0])

        rewards = torch.cat(rewards)
        rewards = self.accelerator.gather(rewards).cpu().numpy()

        self.accelerator.log(
            {
                "reward": rewards,
                "epoch": epoch,
                "reward_mean": rewards.mean(),
                "reward_std": rewards.std(),
            }
        )
        if self.subject_key!="":
            self.accelerator.log({
                f"{self.subject_key}_reward": rewards,
                f"{self.subject_key}_reward_mean": rewards.mean(),
                f"{self.subject_key}_reward_std": rewards.std(),
            })

        if self.config.per_prompt_stat_tracking:
            # gather the prompts across processes
            prompt_ids = self.accelerator.gather(samples["prompt_ids"]).cpu().numpy()
            prompts = self.sd_pipeline.tokenizer.batch_decode(prompt_ids, skip_special_tokens=True)
            advantages = self.stat_tracker.update(prompts, rewards)
        elif normalize_rewards:
            advantages = (rewards - rewards.mean()) / (rewards.std() + 1e-8)
        else:
            advantages=rewards

        # ungather advantages;  keep the entries corresponding to the samples on this process
        samples["advantages"] = (
            torch.as_tensor(advantages)
            .reshape(self.accelerator.num_processes, -1)[self.accelerator.process_index]
            .to(self.accelerator.device)
        )

        del samples["prompt_ids"]

        total_batch_size, num_timesteps = samples["timesteps"].shape

        for inner_epoch in range(self.config.train_num_inner_epochs):
            # shuffle samples along batch dimension
            perm = torch.randperm(total_batch_size, device=self.accelerator.device)
            samples = {k: v[perm] for k, v in samples.items()}

            # shuffle along time dimension independently for each sample
            # still trying to understand the code below
            perms = torch.stack(
                [torch.randperm(num_timesteps, device=self.accelerator.device) for _ in range(total_batch_size)]
            )

            for key in ["timesteps", "latents", "next_latents", "log_probs"]:
                samples[key] = samples[key][
                    torch.arange(total_batch_size, device=self.accelerator.device)[:, None],
                    perms,
                ]

            original_keys = samples.keys()
            original_values = samples.values()
            # rebatch them as user defined train_batch_size is different from sample_batch_size
            reshaped_values = [v.reshape(-1, self.config.train_batch_size, *v.shape[1:]) for v in original_values]

            # Transpose the list of original values
            transposed_values = zip(*reshaped_values)
            # Create new dictionaries for each row of transposed values
            samples_batched = [dict(zip(original_keys, row_values)) for row_values in transposed_values]

            self.sd_pipeline.unet.train()
            global_step = self._train_batched_samples(inner_epoch, epoch, global_step, samples_batched,retain_graph)
            # ensure optimization step at the end of the inner epoch
            if not self.accelerator.sync_gradients:
                raise ValueError(
                    "Optimization step should have been performed by this point. Please check calculated gradient accumulation settings."
                )

        '''if epoch != 0 and epoch % self.config.save_freq == 0 and self.accelerator.is_main_process:
            self.accelerator.save_state()'''

        return global_step

    def calculate_loss(self, latents, timesteps, next_latents, log_probs, advantages, embeds):
        """
        Calculate the loss for a batch of an unpacked sample

        Args:
            latents (torch.Tensor):
                The latents sampled from the diffusion model, shape: [batch_size, num_channels_latents, height, width]
            timesteps (torch.Tensor):
                The timesteps sampled from the diffusion model, shape: [batch_size]
            next_latents (torch.Tensor):
                The next latents sampled from the diffusion model, shape: [batch_size, num_channels_latents, height, width]
            log_probs (torch.Tensor):
                The log probabilities of the latents, shape: [batch_size]
            advantages (torch.Tensor):
                The advantages of the latents, shape: [batch_size]
            embeds (torch.Tensor):
                The embeddings of the prompts, shape: [2*batch_size or batch_size, ...]
                Note: the "or" is because if train_cfg is True, the expectation is that negative prompts are concatenated to the embeds

        Returns:
            loss (torch.Tensor), approx_kl (torch.Tensor), clipfrac (torch.Tensor)
            (all of these are of shape (1,))
        """
        with self.autocast():
            added_cond_kwargs ={}
            if self.use_ip_adapter:
                image_embeds = self.sd_pipeline.sd_pipeline.prepare_ip_adapter_image_embeds(
                    self.ip_adapter_src_image,
                    None,
                    self.sd_pipeline.unet.device,
                    len(latents),
                    self.config.train_cfg,
                )
                added_cond_kwargs = {"image_embeds": image_embeds}

            if self.config.train_cfg:
                noise_pred = self.sd_pipeline.unet(
                    torch.cat([latents] * 2),
                    torch.cat([timesteps] * 2),
                    embeds,
                    added_cond_kwargs=added_cond_kwargs
                ).sample
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + self.config.sample_guidance_scale * (
                    noise_pred_text - noise_pred_uncond
                )
            else:
                noise_pred = self.sd_pipeline.unet(
                    latents,
                    timesteps,
                    embeds,
                    added_cond_kwargs=added_cond_kwargs
                ).sample
            # compute the log prob of next_latents given latents under the current model

            scheduler_step_output = self.sd_pipeline.scheduler_step(
                noise_pred,
                timesteps,
                latents,
                eta=self.config.sample_eta,
                prev_sample=next_latents,
            )

            log_prob = scheduler_step_output.log_probs

        advantages = torch.clamp(
            advantages,
            -self.config.train_adv_clip_max,
            self.config.train_adv_clip_max,
        )

        ratio = torch.exp(log_prob - log_probs) #this is the p_{\theta}(x_{t-1}|x_t,c)/p_{\theta-old}(x_{t-1}|x_t,c)

        loss = self.loss(advantages, self.config.train_clip_range, ratio)

        approx_kl = 0.5 * torch.mean((log_prob - log_probs) ** 2)

        clipfrac = torch.mean((torch.abs(ratio - 1.0) > self.config.train_clip_range).float())

        return loss, approx_kl, clipfrac

    def loss(
        self,
        advantages: torch.Tensor,
        clip_range: float,
        ratio: torch.Tensor,
    ):
        unclipped_loss = -advantages * ratio
        clipped_loss = -advantages * torch.clamp(
            ratio,
            1.0 - clip_range,
            1.0 + clip_range,
        )
        return torch.mean(torch.maximum(unclipped_loss, clipped_loss))

    def _setup_optimizer(self, trainable_layers_parameters):
        if self.config.train_use_8bit_adam:
            import bitsandbytes

            optimizer_cls = bitsandbytes.optim.AdamW8bit
        else:
            optimizer_cls = torch.optim.AdamW

        return optimizer_cls(
            trainable_layers_parameters,
            lr=self.config.train_learning_rate,
            betas=(self.config.train_adam_beta1, self.config.train_adam_beta2),
            weight_decay=self.config.train_adam_weight_decay,
            eps=self.config.train_adam_epsilon,
        )

    def _save_model_hook(self, models, weights, output_dir):
        self.sd_pipeline.save_checkpoint(models, weights, output_dir)
        weights.pop()  # ensures that accelerate doesn't try to handle saving of the model

    def _load_model_hook(self, models, input_dir):
        self.sd_pipeline.load_checkpoint(models, input_dir)
        models.pop()  # ensures that accelerate doesn't try to handle loading of the model

    def _generate_samples(self, iterations, batch_size):
        """
        Generate samples from the model

        Args:
            iterations (int): Number of iterations to generate samples for
            batch_size (int): Batch size to use for sampling

        Returns:
            samples (List[Dict[str, torch.Tensor]]), prompt_image_pairs (List[List[Any]])
        """
        samples = []
        prompt_image_pairs = []
        self.sd_pipeline.unet.eval()

        sample_neg_prompt_embeds = self.neg_prompt_embed.repeat(batch_size, 1, 1)

        for _ in range(iterations):
            prompts, prompt_metadata = zip(*[self.prompt_fn() for _ in range(batch_size)])

            prompt_ids = self.sd_pipeline.tokenizer(
                prompts,
                return_tensors="pt",
                padding="max_length",
                truncation=True,
                max_length=self.sd_pipeline.tokenizer.model_max_length,
            ).input_ids.to(self.accelerator.device)
            prompt_embeds = self.sd_pipeline.text_encoder(prompt_ids)[0]

            with self.autocast():
                if self.use_ip_adapter:
                    sd_output=pipeline_step_ip_adapter(self.sd_pipeline.sd_pipeline,
                                                      prompt_embeds=prompt_embeds,
                        negative_prompt_embeds=sample_neg_prompt_embeds,
                        num_inference_steps=self.config.sample_num_steps,
                        guidance_scale=self.config.sample_guidance_scale,
                        eta=self.config.sample_eta,
                        output_type="pil",
                        height=self.image_size,
                        width=self.image_size,
                        ip_adapter_image=self.ip_adapter_src_image)
                else:
                    sd_output = self.sd_pipeline(
                        prompt_embeds=prompt_embeds,
                        negative_prompt_embeds=sample_neg_prompt_embeds,
                        num_inference_steps=self.config.sample_num_steps,
                        guidance_scale=self.config.sample_guidance_scale,
                        eta=self.config.sample_eta,
                        output_type="pil",
                        height=self.image_size,
                        width=self.image_size
                    )

                images = sd_output.images
                latents = sd_output.latents
                log_probs = sd_output.log_probs

            latents = torch.stack(latents, dim=1)  # (batch_size, num_steps + 1, ...)
            log_probs = torch.stack(log_probs, dim=1)  # (batch_size, num_steps, 1)
            timesteps = self.sd_pipeline.scheduler.timesteps.repeat(batch_size, 1)  # (batch_size, num_steps)

            samples.append(
                {
                    "prompt_ids": prompt_ids,
                    "prompt_embeds": prompt_embeds,
                    "timesteps": timesteps,
                    "latents": latents[:, :-1],  # each entry is the latent before timestep t
                    "next_latents": latents[:, 1:],  # each entry is the latent after timestep t
                    "log_probs": log_probs,
                    "negative_prompt_embeds": sample_neg_prompt_embeds,
                }
            )
            prompt_image_pairs.append([images, prompts, prompt_metadata])

        return samples, prompt_image_pairs

    def _train_batched_samples(self, inner_epoch, epoch, global_step, batched_samples,retain_graph):
        """
        Train on a batch of samples. Main training segment

        Args:
            inner_epoch (int): The current inner epoch
            epoch (int): The current epoch
            global_step (int): The current global step
            batched_samples (List[Dict[str, torch.Tensor]]): The batched samples to train on

        Side Effects:
            - Model weights are updated
            - Logs the statistics to the accelerator trackers.

        Returns:
            global_step (int): The updated global step
        """
        info = defaultdict(list)
        for _i, sample in enumerate(batched_samples):
            if self.config.train_cfg:
                # concat negative prompts to sample prompts to avoid two forward passes
                embeds = torch.cat([sample["negative_prompt_embeds"], sample["prompt_embeds"]])
            else:
                embeds = sample["prompt_embeds"]

            for j in range(self.num_train_timesteps):
                with self.accelerator.accumulate(self.sd_pipeline.unet):
                    loss, approx_kl, clipfrac = self.calculate_loss(
                        sample["latents"][:, j],
                        sample["timesteps"][:, j],
                        sample["next_latents"][:, j],
                        sample["log_probs"][:, j],
                        sample["advantages"],
                        embeds,
                    )
                    info["approx_kl"].append(approx_kl)
                    info["clipfrac"].append(clipfrac)
                    info["loss"].append(loss)

                    self.accelerator.backward(loss,retain_graph=retain_graph)
                    
                    if self.accelerator.sync_gradients:
                        self.accelerator.clip_grad_norm_(
                            self.trainable_layers.parameters()
                            if not isinstance(self.trainable_layers, list)
                            else self.trainable_layers,
                            self.config.train_max_grad_norm,
                        )
                    self.optimizer.step()
                    self.optimizer.zero_grad()

                # Checks if the accelerator has performed an optimization step behind the scenes
                if self.accelerator.sync_gradients:
                    # log training-related stuff
                    info = {k: torch.mean(torch.stack(v)) for k, v in info.items()}
                    info = self.accelerator.reduce(info, reduction="mean")
                    info.update({"epoch": epoch, "inner_epoch": inner_epoch})
                    self.accelerator.log(info)
                    global_step += 1
                    info = defaultdict(list)
        return global_step

    def _config_check(self) -> Tuple[bool, str]:
        samples_per_epoch = (
            self.config.sample_batch_size * self.accelerator.num_processes * self.config.sample_num_batches_per_epoch
        )
        total_train_batch_size = (
            self.config.train_batch_size
            * self.accelerator.num_processes
            * self.config.train_gradient_accumulation_steps
        )

        if not self.config.sample_batch_size >= self.config.train_batch_size:
            return (
                False,
                f"Sample batch size ({self.config.sample_batch_size}) must be greater than or equal to the train batch size ({self.config.train_batch_size})",
            )
        if not self.config.sample_batch_size % self.config.train_batch_size == 0:
            return (
                False,
                f"Sample batch size ({self.config.sample_batch_size}) must be divisible by the train batch size ({self.config.train_batch_size})",
            )
        if not samples_per_epoch % total_train_batch_size == 0:
            return (
                False,
                f"Number of samples per epoch ({samples_per_epoch}) must be divisible by the total train batch size ({total_train_batch_size})",
            )
        return True, ""

    def train(self, epochs: Optional[int] = None,retain_graph:bool=False,normalize_rewards:bool=False):
        """
        Train the model for a given number of epochs
        """
        global_step = 0
        if epochs is None:
            epochs = self.config.num_epochs
        for epoch in range(self.first_epoch, epochs):
            global_step = self.step(epoch, global_step,retain_graph,normalize_rewards)

    def create_model_card(self, path: str, model_name: Optional[str] = "TRL DDPO Model") -> None:
        """Creates and saves a model card for a TRL model.

        Args:
            path (`str`): The path to save the model card to.
            model_name (`str`, *optional*): The name of the model, defaults to `TRL DDPO Model`.
        """
        try:
            user = whoami()["name"]
        # handle the offline case
        except Exception:
            warnings.warn("Cannot retrieve user information assuming you are running in offline mode.")
            return

        if not os.path.exists(path):
            os.makedirs(path)

        model_card_content = MODEL_CARD_TEMPLATE.format(model_name=model_name, model_id=f"{user}/{path}")
        with open(os.path.join(path, "README.md"), "w", encoding="utf-8") as f:
            f.write(model_card_content)

    def _save_pretrained(self, save_directory):
        self.sd_pipeline.save_pretrained(save_directory)
        self.create_model_card(save_directory)


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
import torch

from facenet_pytorch import MTCNN
from PIL import Image
from transformers import Blip2Processor, Blip2ForConditionalGeneration
from src.experiment_helpers.clothing import get_segmentation_model,clothes_segmentation
from src.experiment_helpers.measuring import get_face_caption,get_fashion_caption

device="cpu"
if torch.cuda.is_available():
    device="cuda"
mtcnn=MTCNN(device=device)
segmentation_model=get_segmentation_model(device=device,dtype=torch.float32)


images=[Image.open("ArcaneJinx.jpg"),Image.open("walz.jpg")]

blip_processor = Blip2Processor.from_pretrained("Salesforce/blip2-opt-2.7b")
blip_conditional_gen = Blip2ForConditionalGeneration.from_pretrained("Salesforce/blip2-opt-2.7b").eval()

for image in images:
    face=get_face_caption(image,blip_processor,blip_conditional_gen,mtcnn,10)
    fashion=get_fashion_caption(image,blip_processor,blip_conditional_gen,segmentation_model,0)
    print("face",face)
    print("fashion",fashion)


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.better_ddpo_pipeline import BetterDefaultDDPOStableDiffusionPipeline
from src.experiment_helpers.lora_loading import save_lora_weights, load_lora_weights,save_pipeline_hf,get_pipeline_from_hf

pipeline=BetterDefaultDDPOStableDiffusionPipeline(False,False,True,False,use_lora=True,
    pretrained_model_name="runwayml/stable-diffusion-v1-5")

pipeline("single step",num_inference_steps=1)

hub_model_id="jlbaker361/test-lora-loading"

save_pipeline_hf(pipeline,hub_model_id)
pipeline=get_pipeline_from_hf(hub_model_id,False,False,True,False,use_lora=True,
    pretrained_model_name="runwayml/stable-diffusion-v1-5")

pipeline("single step",num_inference_steps=1)


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.unsafe_stable_diffusion_pipeline import UnsafeStableDiffusionPipeline

pipeline=UnsafeStableDiffusionPipeline.from_pretrained("runwayml/stable-diffusion-v1-5")
pipeline("genital",num_inference_steps=30).images[0].save("nsfw.png")


import torch
from torch import nn
import numpy as np
import torchvision.transforms.functional as F
from PIL import Image
from torchvision.transforms import PILToTensor,Compose,Resize,ToPILImage
from facenet_pytorch import MTCNN, InceptionResnetV1,extract_face

#credits to https://github.com/fdbtrs/ElasticFace/tree/main

__all__ = ['iresnet18', 'iresnet34', 'iresnet50', 'iresnet100']

def rescale_around_zero(image_tensor:torch.Tensor)->torch.Tensor:
    max_value=torch.max(image_tensor)
    min_value=torch.min(image_tensor)
    if max_value >1 and min_value>=0: #in range 0-255
        image_tensor=(image_tensor -128)/128
    elif min_value>=0 and max_value<=1.0: #in range 0-1
        image_tensor=(image_tensor -0.5)/0.5
    return image_tensor

def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=stride,
                     padding=dilation,
                     groups=groups,
                     bias=False,
                     dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=1,
                     stride=stride,
                     bias=False)
class SEModule(nn.Module):
    def __init__(self, channels, reduction):
        super(SEModule, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc1 = nn.Conv2d(channels, channels // reduction, kernel_size=1, padding=0, bias=False)
        self.relu = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv2d(channels // reduction, channels, kernel_size=1, padding=0, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        input = x
        x = self.avg_pool(x)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)

        return input * x

class IBasicBlock(nn.Module):
    expansion = 1
    def __init__(self, inplanes, planes, stride=1, downsample=None,
                 groups=1, base_width=64, dilation=1,use_se=False):
        super(IBasicBlock, self).__init__()
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.bn1 = nn.BatchNorm2d(inplanes, eps=1e-05,)
        self.conv1 = conv3x3(inplanes, planes)
        self.bn2 = nn.BatchNorm2d(planes, eps=1e-05,)
        self.prelu = nn.PReLU(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn3 = nn.BatchNorm2d(planes, eps=1e-05,)
        self.downsample = downsample
        self.stride = stride
        self.use_se=use_se
        if (use_se):
         self.se_block=SEModule(planes,16)

    def forward(self, x):
        identity = x
        out = self.bn1(x)
        out = self.conv1(out)
        out = self.bn2(out)
        out = self.prelu(out)
        out = self.conv2(out)
        out = self.bn3(out)
        if(self.use_se):
            out=self.se_block(out)
        if self.downsample is not None:
            identity = self.downsample(x)
        out += identity
        return out


class IResNet(nn.Module):
    fc_scale = 7 * 7
    def __init__(self,
                 block, layers, dropout=0, num_features=512, zero_init_residual=False,
                 groups=1, width_per_group=64, replace_stride_with_dilation=None, use_se=False):
        super(IResNet, self).__init__()
        self.inplanes = 64
        self.dilation = 1
        self.use_se=use_se
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes, eps=1e-05)
        self.prelu = nn.PReLU(self.inplanes)
        self.layer1 = self._make_layer(block, 64, layers[0], stride=2 ,use_se=self.use_se)
        self.layer2 = self._make_layer(block,
                                       128,
                                       layers[1],
                                       stride=2,
                                       dilate=replace_stride_with_dilation[0],use_se=self.use_se)
        self.layer3 = self._make_layer(block,
                                       256,
                                       layers[2],
                                       stride=2,
                                       dilate=replace_stride_with_dilation[1] ,use_se=self.use_se)
        self.layer4 = self._make_layer(block,
                                       512,
                                       layers[3],
                                       stride=2,
                                       dilate=replace_stride_with_dilation[2] ,use_se=self.use_se)
        self.bn2 = nn.BatchNorm2d(512 * block.expansion, eps=1e-05,)
        self.dropout =nn.Dropout(p=dropout, inplace=True) # 7x7x 512
        self.fc = nn.Linear(512 * block.expansion * self.fc_scale, num_features)
        self.features = nn.BatchNorm1d(num_features, eps=1e-05)
        nn.init.constant_(self.features.weight, 1.0)
        self.features.weight.requires_grad = False

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.normal_(m.weight, 0, 0.1)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, IBasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False,use_se=False):
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                nn.BatchNorm2d(planes * block.expansion, eps=1e-05, ),
            )
        layers = []
        layers.append(
            block(self.inplanes, planes, stride, downsample, self.groups,
                  self.base_width, previous_dilation,use_se=use_se))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(self.inplanes,
                      planes,
                      groups=self.groups,
                      base_width=self.base_width,
                      dilation=self.dilation,use_se=use_se))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.prelu(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.bn2(x)
        x = torch.flatten(x, 1)
        x = self.dropout(x)
        x = self.fc(x)
        x = self.features(x)
        return x


def _iresnet(arch, block, layers, pretrained, progress, **kwargs):
    model = IResNet(block, layers, **kwargs)
    if pretrained:
        raise ValueError()
    return model


def iresnet18(pretrained=False, progress=True, **kwargs):
    return _iresnet('iresnet18', IBasicBlock, [2, 2, 2, 2], pretrained,
                    progress, **kwargs)


def iresnet34(pretrained=False, progress=True, **kwargs):
    return _iresnet('iresnet34', IBasicBlock, [3, 4, 6, 3], pretrained,
                    progress, **kwargs)


def iresnet50(pretrained=False, progress=True, **kwargs):
    return _iresnet('iresnet50', IBasicBlock, [3, 4, 14, 3], pretrained,
                    progress, **kwargs)


def iresnet100(pretrained=False, progress=True, **kwargs):
    return _iresnet('iresnet100', IBasicBlock, [3, 13, 30, 3], pretrained,
                    progress, **kwargs)

class SquarePad:
    def __call__(self, image):
        if isinstance(image,Image.Image):
            w, h = image.size
        elif isinstance(image,torch.Tensor):
            c,w,h=image.shape
        max_wh = np.max([w, h])
        hp = int((max_wh - w) / 2)
        vp = int((max_wh - h) / 2)
        padding = (hp, vp, hp, vp)
        return F.pad(image, padding, 0, 'constant')
    
class CustomPILToTensor:
    def __call__(self, image) -> torch.Tensor:
        if isinstance(image,torch.Tensor):
            return image
        elif isinstance(image,Image.Image):
            return PILToTensor()(image)

def get_iresnet_model(
        device,
        num_features=512,
        weight_path="/scratch/jlb638/faces/elastic_face_arc.pth",
        net=iresnet100
)->IResNet:
    state_dict=torch.load(weight_path,map_location=device)
    model=net(num_features=num_features)
    model.load_state_dict(state_dict)
    model.to(device)
    model.eval()
    return model

def preprocess_iresnet(
  images:list,
  device:str
)->torch.Tensor:
    return_images=[]
    composition=Compose([
        SquarePad(),
        Resize((112,112)),
        CustomPILToTensor()
    ])
    images=[composition(i).to(torch.float32) for i in images]
    images=[rescale_around_zero(i) for i in images]
    #print('processed min max',min_value,max_value)
    return torch.stack(images).to(device)

def get_face_embedding(images:list,mtcnn:MTCNN, iresnet:IResNet,margin:int)->torch.tensor:
    face_tensors=[]
    for img in images:
        boxes, probs=mtcnn.detect(img)
        if boxes is not None:
            extracted_face_tensor=extract_face(img,boxes[0],112,margin)
        else:
            extracted_face_tensor=torch.ones((3,112,112)) #if we dont find any faces :(
        face_tensors.append(extracted_face_tensor)

    iresnet_input=preprocess_iresnet(face_tensors,mtcnn.device)
    return iresnet(iresnet_input)

def face_mask(image:Image.Image,mtcnn:MTCNN,margin:int)->Image.Image:
    image_tensor=PILToTensor()(image)
    mask = torch.zeros_like(image_tensor)
    boxes, probs=mtcnn.detect(image)
    if boxes is not None:
        bbox=boxes[0]
        #print(bbox)
        x_min, y_min, x_max, y_max = bbox
        mask[:, int(y_min)-margin:int(y_max)+margin, int(x_min)-margin:int(x_max)+margin] = 1
    masked_image_tensor = image_tensor * mask
    masked_image=ToPILImage()(masked_image_tensor)
    return masked_image
    



if __name__ == "__main__":
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    weight_path="/scratch/jlb638/faces/elastic_face_arc.pth"
    state_dict=torch.load(weight_path,map_location=device)
    embedding_size=512
    model=iresnet100(num_features=embedding_size)
    model.load_state_dict(state_dict)
    model.eval()
    from PIL import Image
    from torchvision.transforms import PILToTensor,Compose,Resize
    composition=Compose([
        SquarePad(),
        Resize((112,112)),
        PILToTensor()
    ])
    pil_img=Image.open("gnome.jpg")
    '''pt_img=composition(pil_img) #this will be unit8 ~ [0,255] we need to convert it to [-1,1] for iresnet
    pt_img=pt_img.to(torch.float32).unsqueeze(0)
    pt_img= ((pt_img / 255) - 0.5) / 0.5
    embedding=model(pt_img)'''
    images=preprocess_iresnet([pil_img])
    embedding=model(images)
    print(embedding)

    mtcnn = MTCNN()
    embedding=get_face_embedding([pil_img],mtcnn,model,10)
    print(embedding)


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.training import train_unet,train_unet_single_prompt
from PIL import Image
from accelerate import Accelerator
from diffusers import StableDiffusionPipeline
from diffusers.models import AutoencoderKL
import torch
from accelerate import Accelerator
from random import randint
from PIL import Image
from torchvision.transforms import PILToTensor
from transformers import CLIPImageProcessor, CLIPTextModel, CLIPTokenizer, CLIPVisionModelWithProjection
import torch.nn.functional as F
from tqdm.auto import tqdm
from peft import LoraConfig, get_peft_model


def basic_test(use_prior_preservation):
    accelerator=Accelerator(log_with="wandb",gradient_accumulation_steps=4)
    accelerator.init_trackers(project_name="testing_bullshit")
    pipeline=StableDiffusionPipeline.from_pretrained("runwayml/stable-diffusion-v1-5")
    pipeline("do this to instantiate things",num_inference_steps=1)
    pipeline.safety_checker=None
    img=Image.open("ArcaneJinx.jpg").convert("RGB").crop((0,0,64,64)).resize((256,256))
    for model in [pipeline.text_encoder, pipeline.vae,pipeline.unet]:
        model.eval()
        model.requires_grad_(False)
    config = LoraConfig(
            r=4,
            lora_alpha=16,
            target_modules=["to_k", "to_q", "to_v", "to_out.0"],
            lora_dropout=0.0,
            bias="none")
    pipeline.unet = get_peft_model(pipeline.unet, config)
    pipeline.unet.print_trainable_parameters()
    opt=torch.optim.AdamW(
        pipeline.unet.parameters(),
        lr=0.001
    )
    pipeline.unet.to(accelerator.device)
    #pipeline.scheduler.to(accelerator.device)
    opt,pipeline.unet=accelerator.prepare(opt,pipeline.unet)
    #pipeline=train_unet(pipeline,5,[img for _ in range(5)],["{}" for _ in range(5)], opt,use_prior_preservation,"girl",1,1.0,"jinx",accelerator,10,0.1)
    pipeline=train_unet_single_prompt(pipeline,5,[img for _ in range(5)]," {} ",
                            opt,use_prior_preservation,"girl",1,1.0,"jinx",accelerator,10,0.1)
    

if __name__=='__main__':
    
    basic_test(False)
    print("normal worked")
    basic_test(True)


import os

def find_latest_checkpoint(directory,pattern):
    files = os.listdir(directory)
    max_e = -1
    latest_checkpoint = None

    for file in files:
        match = pattern.match(file)
        if match:
            e = int(match.group(1))
            if e > max_e:
                max_e = e
                latest_checkpoint = file

    return latest_checkpoint,max_e


from PIL import Image
import os
import sys
sys.path.append('/home/jlb638/Desktop/package')

from src.experiment_helpers.background import remove_background

def concatenate_images(images, direction='horizontal'):
    # Determine the width and height of the final image
    if direction == 'horizontal':
        total_width = sum(img.width for img in images)
        max_height = max(img.height for img in images)
        new_image = Image.new('RGB', (total_width, max_height))
        
        x_offset = 0
        for img in images:
            new_image.paste(img, (x_offset, 0))
            x_offset += img.width
    elif direction == 'vertical':
        total_height = sum(img.height for img in images)
        max_width = max(img.width for img in images)
        new_image = Image.new('RGB', (max_width, total_height))
        
        y_offset = 0
        for img in images:
            new_image.paste(img, (0, y_offset))
            y_offset += img.height
    else:
        raise ValueError("Direction must be 'horizontal' or 'vertical'")
    
    return new_image


src=Image.open("ArcaneJinx.jpg")
images=[src]
for model_name in  ["u2net", "u2net_human_seg", "u2netp"]:
    dest=remove_background(src,model_name=model_name)
    images.append(dest)
cat=concatenate_images(images)
cat.save("removed.png")


#metrics:
PROMPT_SIMILARITY="prompt_similarity"
IDENTITY_CONSISTENCY="identity_consistency"
TARGET_SIMILARITY="target_similarity"
AESTHETIC_SCORE="aesthetic_score"
IMAGE_REWARD="image_reward"
BLIP_TARGET_CAPTION_SIMILARITY="blip_target_caption_similarity"
BLIP_PROMPT_CAPTION_SIMILARITY="blip_prompt_caption_similarity"
BLIP_TARGET_SIMILARITY="blip_target_similarity"
BLIP_IDENTITY_CONSISTENCY="blip_identity_consistency"
VIT_TARGET_SIMILARITY="vit_target_similarity"
VIT_IDENTITY_CONSISTENCY="vit_identity_consistency"
VIT_STYLE_TARGET_SIMILARITY="style_target_similarity"
VIT_STYLE_CONSISTENCY="style_consistency"
VIT_CONTENT_TARGET_SIMILARITY="content_target_similarity"
VIT_CONTENT_CONSISTENCY="content_consistency"
FACE_TARGET_SIMILARITY="face_target_consistency"
FACE_CONSISTENCY="face_consistency"
FASHION_CONSISTENCY="fashion_consistency"
FASHION_SIMILARITY="fashion_similarity"
DREAM_SIMILARITY="dream_similarity"
DREAM_CONSISTENCY="dream_consistency"

METRIC_LIST=[PROMPT_SIMILARITY, IDENTITY_CONSISTENCY, TARGET_SIMILARITY, AESTHETIC_SCORE, IMAGE_REWARD,
    BLIP_TARGET_CAPTION_SIMILARITY,BLIP_PROMPT_CAPTION_SIMILARITY,
    #BLIP_TARGET_SIMILARITY,BLIP_IDENTITY_CONSISTENCY,
    VIT_TARGET_SIMILARITY,VIT_IDENTITY_CONSISTENCY,
    VIT_STYLE_TARGET_SIMILARITY, VIT_STYLE_CONSISTENCY,
    VIT_CONTENT_TARGET_SIMILARITY,VIT_CONTENT_CONSISTENCY,
    FACE_TARGET_SIMILARITY, FACE_CONSISTENCY,FASHION_CONSISTENCY,FASHION_SIMILARITY,
    DREAM_SIMILARITY, DREAM_CONSISTENCY]

FACE_METRIC_LIST=[FACE_TARGET_SIMILARITY, FACE_CONSISTENCY]


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')

from src.experiment_helpers.measuring import get_metric_dict
from PIL import Image

images=[Image.open("ArcaneJinx.jpg"),Image.open("nsfw.png")]

metric_dict=get_metric_dict(["girl","person"],images,images)
for k,v in metric_dict.items():
    print(k,v)


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
import torch
from PIL import Image
from datasets import load_dataset
import torch
from torchvision.transforms import PILToTensor

from src.experiment_helpers.cloth_process import check_or_download_model,load_seg_model,get_palette,generate_mask
from src.experiment_helpers.clothing import get_segmentation_model,get_mask,clothes_segmentation


checkpoint="/scratch/jlb638/fashion_segmentation/cloth_segm.pth"

check_or_download_model(checkpoint)

device="cpu"

palette=get_palette(4)

net = load_seg_model(checkpoint, device=device)
for i,row in enumerate(load_dataset("jlbaker361/humans",split="train")):
    image=row["splash"] #Image.open("ArcaneJinx.jpg").convert("RGB")
    cloth_seg = generate_mask(image, net=net, device=device)
    segmentation_model=get_segmentation_model(device,torch.float32)
    '''tensor_img=PILToTensor()(image)
    tensor_img=tensor_img.to(torch.float32)
    cloth_seg=clothes_segmentation(image,segmentation_model,0)'''
    cloth_seg.save(f"segmented_{i}.jpg")
    if i>5:
        break
    #cloth_seg.save("masked_jinx.jpg")


from accelerate import Accelerator
from .static_globals import *
from .better_vit_model import BetterViTModel
import ImageReward as image_reward
import torch

from transformers import CLIPProcessor, CLIPModel,ViTImageProcessor, ViTModel
import numpy as np
from numpy.linalg import norm
from .aesthetic_reward import AestheticScorer
from transformers import Blip2Processor, Blip2ForConditionalGeneration
from PIL import Image
from sentence_transformers import SentenceTransformer
from sentence_transformers.util import cos_sim as cos_sim_st
from facenet_pytorch import MTCNN
from .elastic_face_iresnet import get_face_embedding,get_iresnet_model,face_mask
from .cloth_network import U2NET
from .cloth_process import load_seg_model,get_palette,generate_mask
import wandb
import random, string
from .legacy_dreamsim import dreamsim


def randomword(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))

def cos_sim(vector_i,vector_j)->float:
    if type(vector_i)==torch.Tensor:
        vector_i=vector_i.cpu().detach()
    if type(vector_j)==torch.Tensor:
        vector_j=vector_j.cpu().detach()
    return np.dot(vector_i,vector_j)/(norm(vector_i)*norm(vector_j))

def get_caption(image:Image,blip_processor: Blip2Processor,blip_conditional_gen: Blip2ForConditionalGeneration):
    caption_inputs = blip_processor(image, "", return_tensors="pt")
    for name in ["pixel_values","input_ids"]:
        caption_inputs[name]=caption_inputs[name].to(blip_conditional_gen.device)
    caption_out=blip_conditional_gen.generate(**caption_inputs)
    return blip_processor.decode(caption_out[0],skip_special_tokens=True).strip()

def get_face_caption(image:Image,blip_processor: Blip2Processor,blip_conditional_gen: Blip2ForConditionalGeneration,mtcnn:MTCNN,margin:int)->str:
    face_image=face_mask(image,mtcnn,margin)
    #face_image.save(randomword(3)+"_face.jpg")
    return get_caption(face_image, blip_processor, blip_conditional_gen)

def get_fashion_caption(image:Image,blip_processor: Blip2Processor,blip_conditional_gen: Blip2ForConditionalGeneration,seg_model:U2NET)->str:
    #fashion_image=clothes_segmentation(image,segmentation_model,threshold)
    fashion_image=generate_mask(image,seg_model,blip_conditional_gen.device)
    #fashion_image.save(randomword(3)+"_fashion.jpg")
    return get_caption(fashion_image, blip_processor, blip_conditional_gen)

def get_vit_embeddings(vit_processor: ViTImageProcessor, vit_model: BetterViTModel, image_list:list,return_numpy:bool=True):
    vit_embedding_list=[]
    vit_content_embedding_list=[]
    vit_style_embedding_list=[]
    for image in image_list:
        vit_inputs = vit_processor(images=[image], return_tensors="pt")
        #print("inputs :)")
        vit_inputs['pixel_values']=vit_inputs['pixel_values'].to(vit_model.device)
        vit_outputs=vit_model(**vit_inputs,output_hidden_states=True, output_past_key_values=True)
        vit_embedding_list.append(vit_outputs.last_hidden_state.reshape(1,-1)[0])
        vit_style_embedding_list.append(vit_outputs.last_hidden_state[0][0]) #CLS token: https://github.com/google/dreambooth/issues/3
        vit_content_embedding_list.append(vit_outputs.past_key_values[11][0].reshape(1,-1)[0])
    if return_numpy:
        vit_embedding_list=[v.cpu().numpy() for v in vit_embedding_list]
        vit_style_embedding_list=[v.cpu().numpy() for v in vit_style_embedding_list]
        vit_content_embedding_list=[v.cpu().numpy() for v in vit_content_embedding_list]
    return vit_embedding_list,vit_style_embedding_list, vit_content_embedding_list

@torch.no_grad()
def get_metric_dict(evaluation_prompt_list:list, evaluation_image_list:list,src_image_list:list,accelerator:Accelerator=None,
                    use_face:bool=False,)->dict:
    metric_dict={}
    
    clip_model = CLIPModel.from_pretrained("openai/clip-vit-base-patch32").eval()
    clip_processor = CLIPProcessor.from_pretrained("openai/clip-vit-base-patch32")
    
    if accelerator is not None:
        clip_model.to(accelerator.device)
        clip_model=accelerator.prepare(clip_model)
    evaluation_image_embed_list=[]
    evaluation_image_fashion_embed_list=[]
    text_embed_list=[]
    src_image_embed_list=[]
    src_image_fashion_embed_list=[]
    for images,image_embed_list in zip([evaluation_image_list, src_image_list], [evaluation_image_embed_list, src_image_embed_list]):
        for image in images:
            clip_inputs=clip_processor(text=[" "], images=[image], return_tensors="pt", padding=True)
            clip_inputs["input_ids"]=clip_inputs["input_ids"].to(clip_model.device)
            clip_inputs["pixel_values"]=clip_inputs["pixel_values"].to(clip_model.device)
            clip_inputs["attention_mask"]=clip_inputs["attention_mask"].to(clip_model.device)
            try:
                clip_inputs["position_ids"]= clip_inputs["position_ids"].to(clip_model.device)
            except:
                pass

            clip_outputs = clip_model(**clip_inputs)
            image_embed_list.append(clip_outputs.image_embeds.detach().cpu().numpy()[0])

    clip_inputs=clip_processor(text=evaluation_prompt_list, images=[image], return_tensors="pt", padding=True)
    clip_inputs["input_ids"]=clip_inputs["input_ids"].to(clip_model.device)
    clip_inputs["pixel_values"]=clip_inputs["pixel_values"].to(clip_model.device)
    clip_inputs["attention_mask"]=clip_inputs["attention_mask"].to(clip_model.device)
    try:
        clip_inputs["position_ids"]= clip_inputs["position_ids"].to(clip_model.device)
    except:
        pass

    clip_outputs = clip_model(**clip_inputs)
    text_embed_list=clip_outputs.text_embeds.cpu().detach().numpy()

    fashion_clip_processor=CLIPProcessor.from_pretrained("patrickjohncyh/fashion-clip")
    fashion_clip_model=CLIPModel.from_pretrained("patrickjohncyh/fashion-clip").eval()

    for images,image_embed_list in zip([evaluation_image_list, src_image_list], [evaluation_image_fashion_embed_list, src_image_fashion_embed_list]):
        for image in images:
            fashion_clip_inputs=fashion_clip_processor(text=[" "], images=[image], return_tensors="pt", padding=True)
            fashion_clip_inputs["input_ids"]=fashion_clip_inputs["input_ids"].to(fashion_clip_model.device)
            fashion_clip_inputs["pixel_values"]=fashion_clip_inputs["pixel_values"].to(fashion_clip_model.device)
            fashion_clip_inputs["attention_mask"]=fashion_clip_inputs["attention_mask"].to(fashion_clip_model.device)
            try:
                fashion_clip_inputs["position_ids"]= fashion_clip_inputs["position_ids"].to(fashion_clip_model.device)
            except:
                pass

            fashion_clip_outputs = fashion_clip_model(**fashion_clip_inputs)
            fashion_embedding=fashion_clip_outputs.image_embeds.detach().cpu().numpy()[0]
            image_embed_list.append(fashion_embedding)
    
    
    device="cpu"
    if accelerator is not None:
        device=accelerator.device
    ir_model=image_reward.load("/scratch/jlb638/reward-blob",med_config="/scratch/jlb638/ImageReward/med_config.json",device=device)

    identity_consistency_list=[]
    target_similarity_list=[]
    prompt_similarity_list=[]
    for i in range(len(evaluation_image_embed_list)):
        image_embed=evaluation_image_embed_list[i]
        text_embed=text_embed_list[i]
        for src_image_embed in src_image_embed_list:
            target_similarity_list.append(cos_sim(image_embed,src_image_embed))
        prompt_similarity_list.append(cos_sim(image_embed, text_embed))
        for j in range(i+1, len(evaluation_image_embed_list)):
            #print(i,j)
            vector_j=evaluation_image_embed_list[j]
            sim=cos_sim(image_embed,vector_j)
            identity_consistency_list.append(sim)

    fashion_consistency_list=[]
    fashion_similarity_list=[]
    for i in range(len(evaluation_image_fashion_embed_list)):
        fashion_image_embed=evaluation_image_fashion_embed_list[i]
        for fashion_src_embed in src_image_fashion_embed_list:
            fashion_similarity_list.append(cos_sim(fashion_image_embed, fashion_src_embed))
        for j in range(i+1, len(evaluation_image_fashion_embed_list) ):
            fashion_consistency_list.append(cos_sim(fashion_image_embed,evaluation_image_fashion_embed_list[j] ))

    metric_dict[IDENTITY_CONSISTENCY]=np.mean(identity_consistency_list)
    metric_dict[TARGET_SIMILARITY]=np.mean(target_similarity_list)
    metric_dict[PROMPT_SIMILARITY]=np.mean(prompt_similarity_list)

    metric_dict[FASHION_CONSISTENCY]=np.mean(fashion_consistency_list)
    metric_dict[FASHION_SIMILARITY]=np.mean(fashion_similarity_list)

    '''blip_processor = Blip2Processor.from_pretrained("Salesforce/blip2-opt-2.7b")
    blip_conditional_gen = Blip2ForConditionalGeneration.from_pretrained("Salesforce/blip2-opt-2.7b").eval()
    if accelerator is not None:
        blip_conditional_gen.to(accelerator.device)
        blip_conditional_gen=accelerator.prepare(blip_conditional_gen)'''
    

    '''src_blip_caption_list=[get_caption(src_image,blip_processor,blip_conditional_gen) for src_image in src_image_list]
    image_blip_caption_list=[get_caption(image,blip_processor,blip_conditional_gen) for image in evaluation_image_list]'''
    embedding_model = SentenceTransformer('Alibaba-NLP/gte-large-en-v1.5', trust_remote_code=True).eval()
    if accelerator is not None:
        embedding_model.to(accelerator.device)
        embedding_model=accelerator.prepare(embedding_model)
    '''src_blip_embedding_list=embedding_model.encode(src_blip_caption_list)
    image_blip_embedding_list=embedding_model.encode(image_blip_caption_list)
    evaluation_blip_embedding_list=embedding_model.encode(evaluation_prompt_list)'''


    '''metric_dict[BLIP_TARGET_CAPTION_SIMILARITY]=np.mean(cos_sim_st(src_blip_embedding_list, image_blip_embedding_list).cpu().detach().numpy())
    metric_dict[BLIP_PROMPT_CAPTION_SIMILARITY]=np.mean(
        [cos_sim_st(evaluation_blip_embedding, image_blip_embedding).cpu().detach().numpy() for evaluation_blip_embedding, image_blip_embedding in zip(evaluation_blip_embedding_list, image_blip_embedding_list)]
    )'''

    '''blip_model=Blip2Model.from_pretrained("Salesforce/blip2-opt-2.7b")
    blip_inputs=blip_processor(text=evaluation_prompt_list, images=evaluation_image_list+image_list)
    blip_outputs=blip_model(**blip_inputs)'''

    vit_processor=ViTImageProcessor.from_pretrained("facebook/dino-vits16")
    vit_model=BetterViTModel.from_pretrained("facebook/dino-vits16").eval()
    if accelerator is not None:
        vit_model.to(accelerator.device)
        vit_model=accelerator.prepare(vit_model)

    image_vit_embedding_list,image_vit_style_embedding_list,image_vit_content_embedding_list=get_vit_embeddings(vit_processor,vit_model,src_image_list)
    evaluation_image_vit_embedding_list,evaluation_image_vit_style_embedding_list,evaluation_image_vit_content_embedding_list=get_vit_embeddings(vit_processor,vit_model,evaluation_image_list)

    vit_consistency_list=[]
    vit_target_similarity_list=[]
    vit_style_consistency_list=[]
    vit_style_target_similarity_list=[]
    vit_content_consistency_list=[]
    vit_content_similarity_list=[]

    for i in range(len(evaluation_image_list)):
        evaluation_vit_embedding=evaluation_image_vit_embedding_list[i]
        evaluation_vit_content_embedding=evaluation_image_vit_content_embedding_list[i]
        evaluation_vit_style_embedding=evaluation_image_vit_style_embedding_list[i]
        for k in range(len(src_image_list)):
            image_vit_embedding=image_vit_embedding_list[k]
            image_vit_style_embedding=image_vit_style_embedding_list[k]
            image_vit_content_embedding=image_vit_content_embedding_list[k]
            vit_target_similarity_list.append(cos_sim(image_vit_embedding, evaluation_vit_embedding))
            vit_style_target_similarity_list.append(cos_sim(evaluation_vit_style_embedding,image_vit_style_embedding))
            vit_content_similarity_list.append(cos_sim(evaluation_vit_content_embedding,image_vit_content_embedding))
        for j in range(i+1, len(evaluation_image_list)):
            secondary_evaluation_vit_embedding=evaluation_image_vit_embedding_list[j]
            secondary_evaluation_vit_content_embedding=evaluation_image_vit_content_embedding_list[j]
            secondary_evaluation_vit_style_embedding=evaluation_image_vit_style_embedding_list[j]
            vit_consistency_list.append(cos_sim(evaluation_vit_embedding,secondary_evaluation_vit_embedding))
            vit_style_consistency_list.append(cos_sim(evaluation_vit_style_embedding, secondary_evaluation_vit_style_embedding))
            vit_content_consistency_list.append(cos_sim(evaluation_vit_content_embedding,secondary_evaluation_vit_content_embedding))

    metric_dict[VIT_TARGET_SIMILARITY]=np.mean(vit_target_similarity_list)
    metric_dict[VIT_IDENTITY_CONSISTENCY]=np.mean(vit_consistency_list)
    metric_dict[VIT_STYLE_TARGET_SIMILARITY]=np.mean(vit_style_target_similarity_list)
    metric_dict[VIT_STYLE_CONSISTENCY]=np.mean(vit_style_consistency_list)
    metric_dict[VIT_CONTENT_TARGET_SIMILARITY]=np.mean(vit_content_similarity_list)
    metric_dict[VIT_CONTENT_CONSISTENCY]=np.mean(vit_content_consistency_list)

    metric_dict[IMAGE_REWARD]=np.mean(
        [ir_model.score(evaluation_prompt,evaluation_image) for evaluation_prompt,evaluation_image in zip(evaluation_prompt_list, evaluation_image_list) ]
    )
    aesthetic_scorer=AestheticScorer()
    if accelerator is not None:
        aesthetic_scorer.to(accelerator.device)
        aesthetic_scorer=accelerator.prepare(aesthetic_scorer)
    metric_dict[AESTHETIC_SCORE]=np.mean(
        [aesthetic_scorer(evaluation_image).cpu().numpy()[0] for evaluation_image in evaluation_image_list]
    )

    if accelerator is not None:
        accelerator.free_memory()

    torch.cuda.empty_cache()

    dream_model, dream_preprocess = dreamsim(pretrained=True,cache_dir="/scratch/jlb638/dreamsim",device=device)
    evaluation_image_dream_embed_list=[]
    src_image_dream_embed_list=[]
    for images,image_embed_list in zip([evaluation_image_list, src_image_list], [evaluation_image_dream_embed_list, src_image_dream_embed_list]):
        for image in images:
            preprocessed=dream_preprocess(image)
            embedding=dream_model.embed(preprocessed)[0]
            image_embed_list.append(embedding)

    dream_consistency_list=[]
    dream_similarity_list=[]

    for i in range(len(evaluation_image_list)):
        embedding=evaluation_image_dream_embed_list[i]
        for k in range(len(src_image_list)):
            src_embedding=src_image_dream_embed_list[k]
            dream_similarity_list.append(cos_sim(embedding,src_embedding))
        for j in range(i+1, len(evaluation_image_list)):
            dream_consistency_list.append(cos_sim(embedding, evaluation_image_dream_embed_list[j]))
    
    metric_dict[DREAM_CONSISTENCY]=np.mean(dream_consistency_list)
    metric_dict[DREAM_SIMILARITY]=np.mean(dream_similarity_list)



    if use_face:
        '''if accelerator is not None:
            mtcnn=MTCNN(device=accelerator.device)
            iresnet=get_iresnet_model(accelerator.device)
            mtcnn,iresnet=accelerator.prepare(mtcnn,iresnet)
        else:'''
        mtcnn=MTCNN(device="cpu") #thinking that maybe these get too big for most GPUs???
        iresnet=get_iresnet_model("cpu")
        mtcnn.requires_grad_(False)
        mtcnn.eval()
        image_face_embedding_list=[get_face_embedding([evaluation_image],mtcnn,iresnet,10)[0] for evaluation_image in evaluation_image_list]
        src_face_embedding_list=[get_face_embedding([src_image],mtcnn,iresnet,10)[0] for src_image in src_image_list]
        face_consistency_list=[]
        face_target_similarity_list=[]
        for i,face_embedding in enumerate(image_face_embedding_list):
            for j,secondary_embedding in enumerate(image_face_embedding_list[i+1:]):
                face_consistency_list.append(cos_sim(secondary_embedding,face_embedding))
            for k,src_face_embedding in enumerate(src_face_embedding_list):
                face_target_similarity_list.append(cos_sim(face_embedding,src_face_embedding))
        metric_dict[FACE_CONSISTENCY]=np.mean(face_consistency_list)
        metric_dict[FACE_TARGET_SIMILARITY]=np.mean(face_target_similarity_list)

    for metric in METRIC_LIST:
        if metric not in metric_dict:
            metric_dict[metric]=0.0
    return metric_dict

def print_tables(run_id_dict:dict,key_list:list,project:str )->None:
    api=wandb.Api(timeout=60)
    print(" & ".join(key_list),"\\\\")
    for name,run_id_list in run_id_dict.items():
        metric_dict={
            key:[] for key in key_list
        }
        for run_id in run_id_list:
            try:
                run=api.run(f"jlbaker361/{project}/{run_id}")
                for key in key_list:
                    try:
                        history=run.history(keys=[key])
                        metric_dict[key].append(history[key][0])
                    except:
                        print(f"couldnt find key {key} for name {name} and {run_id}")
            except:
                pass
        #print(metric_dict)
        print(name.replace("_", " "), "&", " & ".join([f"{round(np.mean(metric_dict[key]),4)}" for key in metric_dict.keys()]),"\\\\")


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.measuring import print_tables


run_id_dict={
        "Vanilla":["173w2r1q","mwhnxtpb"],
        #"Vanilla_Negative":["co78j4ak","ykumm85l"],
        "Vanilla_Spare":["peyzw3oa","07n5xoqb"],
        "T5_Transformer":["3zl9cz3a","g1trkmly","bza128pk","osadows9","3d9go23s","93f0ngw2"],
        #"T5_Transformer_Negative":["y42whkb3","njww0sxj"],
        "T5_Transformer_Spare":["8x7mwl5r","jfeq3nq6","3x2xbzuc","bq1224jb","8ba3jea3","letuoqqs"],
        "T5_Unet": ["yob4st69","qfm2nxy7"],
        #"T5_UNet_Negative":["4yegkplc","1571dfbf"],
        "T5_UNet_Spare":["7a3eacrp","0qetqpq0"],
        "Vanilla_Short":["gd0jkf6f","0j5a8ixd"], #this is the shit in the normal script
        "Vanilla_Short_Spare":["fjlumnap","rmebldhu"],
        "Llama_Unet":["vd003zab","pwugepr3","ru7wudup","bxr7owkj","wba5xtkj","ecgcazyt"],
        #"Llama_UNet_Negative":[],
        "Llama_Unet_Spare":["d4t21b4k","wmerq6if","bjqwvq40","enru9x0c","bmdlnj49","adeyybau"]
    }
key_list=["content_target_similarity", "content_consistency", "image_reward", "prompt_similarity"]

print_tables(run_id_dict,key_list,"personalization")


from diffusers import StableDiffusionPipeline
from diffusers.models import AutoencoderKL
import torch
from accelerate import Accelerator
from random import randint
from PIL import Image
from torchvision.transforms import PILToTensor
from transformers import CLIPImageProcessor, CLIPTextModel, CLIPTokenizer, CLIPVisionModelWithProjection
import torch.nn.functional as F
from tqdm.auto import tqdm
import time
import wandb

def pil_to_tensor_process(image:Image):
    tensor=PILToTensor()(image)
    if torch.max(tensor)>1.0 and torch.min(tensor)>=0:
        tensor=tensor/128.0 -1.0
    elif torch.min(tensor)>=0:
        tensor=tensor*2.0 -1.0
    if tensor.ndim==4:
        tensor=tensor[0]
    return tensor

def encode_prompt(text_encoder: CLIPTextModel,
        tokenizer: CLIPTokenizer,prompt:str):
    text_inputs = tokenizer(
                prompt,
                padding="max_length",
                max_length=tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )
    text_input_ids = text_inputs.input_ids
    if hasattr(text_encoder.config, "use_attention_mask") and text_encoder.config.use_attention_mask:
        attention_mask = text_inputs.attention_mask.to(text_encoder.device)
    else:
        attention_mask = None

    prompt_embeds = text_encoder(text_input_ids.to(text_encoder.device), attention_mask=attention_mask)
    #print(prompt_embeds)
    #print(dir(prompt_embeds))
    #secondary_prompt_embeds=text_encoder(text_input_ids.to(text_encoder.device), attention_mask=attention_mask,return_dict=False)
    #print(secondary_prompt_embeds)
    #prompt_embeds = prompt_embeds[0]
    return prompt_embeds.last_hidden_state

def train_unet(pipeline:StableDiffusionPipeline,
                   epochs:int,
                   training_image_list:list,
                   training_prompt_list:list,
                   optimizer:torch.optim.Optimizer,
                   use_prior_preservation:bool,
                   prior_class:str,
                   batch_size:int,
                   max_grad_norm:float,
                   entity_name:str,
                   accelerator:Accelerator,
                   num_inference_steps:int,
                   prior_loss_weight:float,
                   track_loss:bool=False,
                   ip_adapter_image:Image.Image=None)->StableDiffusionPipeline:
    '''
    prior_class= "cat" or "man" or whatever
    training_prompt_list= ["{} doing a thing"] will be formatted with entity_name or if 
        using prior entity name + prior class and prior class
    '''
    
    pipeline.safety_checker=None
    unet=pipeline.unet
    vae=pipeline.vae
    text_encoder=pipeline.text_encoder
    tokenizer=pipeline.tokenizer
    noise_scheduler=pipeline.scheduler

    _training_image_list=[i for i in training_image_list]
    _training_prompt_list=[p for p in training_prompt_list]
    index=0
    while len(_training_image_list)%batch_size!=0:
        index=index%len(training_image_list)
        _training_image_list.append(training_image_list[index])
        _training_prompt_list.append(training_prompt_list[index])
        index+=1
    training_prompt_list=_training_prompt_list
    training_image_list=_training_image_list

    initial_text_encoder_device=text_encoder.device
    text_encoder=text_encoder.to(unet.device)
    initial_vae_device=vae.device
    vae=vae.to(unet.device)
    pipeline("do this to instantiate things",num_inference_steps=1,ip_adapter_image=ip_adapter_image)
    added_cond_kwargs={}
    if ip_adapter_image is not None:
        added_cond_kwargs["image_embeds"]=pipeline.prepare_ip_adapter_image_embeds(ip_adapter_image,None,unet.device,1,False)
    width,height=training_image_list[0].size
    if use_prior_preservation:
        #prior_prompt_list=[p.format(prior_class) for p in training_prompt_list]
        prior_prompt_list=[encode_prompt(text_encoder,tokenizer,p.format(prior_class)) for p in training_prompt_list]
        training_prompt_list=[encode_prompt(text_encoder,tokenizer,p.format(f"{entity_name} {prior_class}")) for p in training_prompt_list]
        prior_image_list=[
            pil_to_tensor_process(pipeline(prompt_embeds=prompt_embeds,num_inference_steps=num_inference_steps,safety_checker=None,width=width,height=height).images[0])
            for prompt_embeds in prior_prompt_list
        ]
        training_image_list=[
            pil_to_tensor_process(image) for image in training_image_list
        ]
        training_image_list_batched=[
            training_image_list[i:i+batch_size]+prior_image_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
        training_prompt_list_batched=[
            training_prompt_list[i:i+batch_size]+prior_prompt_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
    else:
        training_image_list=[
            pil_to_tensor_process(image) for image in training_image_list
        ]
        try:
            training_prompt_list=[encode_prompt(text_encoder,tokenizer,p.format(f"{entity_name}")) for p in training_prompt_list]
        except:
            text_encoder=text_encoder.to(initial_text_encoder_device)
            training_prompt_list=[encode_prompt(text_encoder,tokenizer,p.format(f"{entity_name}")) for p in training_prompt_list]
        training_image_list_batched=[
            training_image_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
        training_prompt_list_batched=[
            training_prompt_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
    text_encoder=text_encoder.to(initial_text_encoder_device)
    vae=vae.to(initial_vae_device)
    training_image_data=[
        torch.stack(batch) for batch in training_image_list_batched
    ]
    training_prompts_data=[
        torch.cat(batch) for batch in training_prompt_list_batched
    ]

    progress_bar = tqdm(
        range(0, epochs),
        initial=0,
        desc="epochs",
        # Only show the progress bar once on each machine.
        disable=not accelerator.is_local_main_process,
    )
    '''print('unet device',unet.device)
    print('vae device',vae.device)
    print('text encoder device',text_encoder.device )'''
    start=time.time()
    for e in range(epochs):
        unet.train()
        for images,prompts in zip(training_image_data, training_prompts_data):
            with accelerator.accumulate(unet):
                '''if use_prior_preservation:
                    images,prior_images=images.chunk(2)
                    prompts,prior_prompts=images.chunk(2)'''
                images=images.to(vae.device)
                model_input = vae.encode(images).latent_dist.sample()
                model_input = model_input * vae.config.scaling_factor
                model_input=model_input.to(unet.device)
                prompts=prompts.to(unet.device)

                noise = torch.randn_like(model_input)
                bsz, channels, height, width = model_input.shape
                # Sample a random timestep for each image
                timesteps = torch.randint(
                    0, noise_scheduler.config.num_train_timesteps, (bsz,), device=unet.device
                )
                timesteps = timesteps.long()

                # Add noise to the model input according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                noisy_model_input = noise_scheduler.add_noise(model_input, noise, timesteps)
                noisy_model_input=noisy_model_input.to(accelerator.device)
                '''print('unet device',unet.device)
                print('noisy_model_input.device',noisy_model_input.device)
                print('timesteps.device',timesteps.device)
                print('prompts.device',prompts.device)'''
                model_pred = unet(
                    noisy_model_input, timesteps, prompts, return_dict=False,added_cond_kwargs=added_cond_kwargs
                )[0]
                if use_prior_preservation:
                    # Chunk the noise and model_pred into two parts and compute the loss on each part separately.
                    model_pred, model_pred_prior = torch.chunk(model_pred, 2, dim=0)
                    noise, noise_prior = torch.chunk(noise, 2, dim=0)
                    # Compute prior loss
                    prior_loss = F.mse_loss(model_pred_prior.float(), noise_prior.float(), reduction="mean")
                loss = F.mse_loss(model_pred.float(), noise.float(), reduction="mean")
                if use_prior_preservation:
                    loss=loss+prior_loss_weight*prior_loss

                accelerator.backward(loss)
                if accelerator.sync_gradients:
                    params_to_clip = (
                        unet.parameters()
                    )
                    accelerator.clip_grad_norm_(params_to_clip, max_grad_norm)
                    if track_loss:
                        try:
                            accelerator.log({
                                "loss":loss
                            })
                        except:
                            accelerator.log({
                                "loss":loss.detach().cpu().numpy()
                            })
                optimizer.step()
                optimizer.zero_grad()
        if accelerator.sync_gradients:
            progress_bar.update(1)
    end=time.time()
    print(f"trained unet! time elapsed: {end-start}")
    return pipeline


def train_unet_single_prompt(pipeline:StableDiffusionPipeline,
                   epochs:int,
                   training_image_list:list,
                   training_prompt:str,
                   optimizer:torch.optim.Optimizer,
                   use_prior_preservation:bool,
                   prior_class:str,
                   batch_size:int,
                   max_grad_norm:float,
                   entity_name:str,
                   accelerator:Accelerator,
                   num_inference_steps:int,
                   prior_loss_weight:float,
                   track_loss:bool=False,
                   n_prior:int=5,
                   log_images:int=-1)->StableDiffusionPipeline:
    '''
    prior_class= "cat" or "man" or whatever
    training_prompt_list= ["{} doing a thing"] will be formatted with entity_name or if 
        using prior entity name + prior class and prior class
    '''
    
    pipeline.safety_checker=None
    unet=pipeline.unet
    vae=pipeline.vae
    text_encoder=pipeline.text_encoder
    tokenizer=pipeline.tokenizer
    noise_scheduler=pipeline.scheduler

    _training_image_list=[i for i in training_image_list]
    index=0
    while len(_training_image_list)%batch_size!=0:
        index=index%len(training_image_list)
        _training_image_list.append(training_image_list[index])
        index+=1
    training_image_list=_training_image_list

    initial_text_encoder_device=text_encoder.device
    text_encoder=text_encoder.to(unet.device)
    initial_vae_device=vae.device
    vae=vae.to(unet.device)
    pipeline("do this to instantiate things",num_inference_steps=1)
    width,height=training_image_list[0].size
    if use_prior_preservation:
        #prior_prompt_list=[p.format(prior_class) for p in training_prompt_list]
        prior_prompt_list=[encode_prompt(text_encoder,tokenizer,training_prompt.format(prior_class)) for _ in range(n_prior)]
        prompts=[encode_prompt(text_encoder,tokenizer,training_prompt.format(f"{entity_name} {prior_class}")) for _ in range(batch_size)]
        prior_image_list=[
            pil_to_tensor_process(pipeline(prompt_embeds=prompt_embeds,num_inference_steps=num_inference_steps,safety_checker=None,width=width,height=height).images[0])
            for prompt_embeds in prior_prompt_list
        ]
        training_image_list=[
            pil_to_tensor_process(image) for image in training_image_list
        ]
        training_image_list_batched=[
            training_image_list[i:i+batch_size]+prior_image_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
    else:
        training_image_list=[
            pil_to_tensor_process(image) for image in training_image_list
        ]
        try:
            prompts=[encode_prompt(text_encoder,tokenizer,training_prompt.format(f"{entity_name}")) for _ in range(batch_size)]
        except:
            text_encoder=text_encoder.to(initial_text_encoder_device)
            prompts=[encode_prompt(text_encoder,tokenizer,training_prompt.format(f"{entity_name}")) for _ in range(batch_size)]
        training_image_list_batched=[
            training_image_list[i:i+batch_size] for i in range(0,len(training_image_list),batch_size)
        ]
    text_encoder=text_encoder.to(initial_text_encoder_device)
    vae=vae.to(initial_vae_device)
    training_image_data=[
        torch.stack(batch) for batch in training_image_list_batched
    ]
    prompts=torch.cat(prompts)
    progress_bar = tqdm(
        range(0, epochs),
        initial=0,
        desc="epochs",
        # Only show the progress bar once on each machine.
        disable=not accelerator.is_local_main_process,
    )
    '''print('unet device',unet.device)
    print('vae device',vae.device)
    print('text encoder device',text_encoder.device )'''
    start=time.time()
    for e in range(epochs):
        unet.train()
        for images in training_image_data:
            with accelerator.accumulate(unet):
                '''if use_prior_preservation:
                    images,prior_images=images.chunk(2)
                    prompts,prior_prompts=images.chunk(2)'''
                images=images.to(vae.device)
                model_input = vae.encode(images).latent_dist.sample()
                model_input = model_input * vae.config.scaling_factor
                model_input=model_input.to(unet.device)
                prompts=[encode_prompt(text_encoder,tokenizer,training_prompt.format(f"{entity_name}")) for _ in range(batch_size)]
                prompts=torch.cat(prompts)
                prompts=prompts.to(unet.device)

                noise = torch.randn_like(model_input,device=unet.device)
                bsz, channels, height, width = model_input.shape
                # Sample a random timestep for each image
                timesteps = torch.randint(
                    0, noise_scheduler.config.num_train_timesteps, (bsz,), device=unet.device
                )
                timesteps = timesteps.long()

                # Add noise to the model input according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                noisy_model_input = noise_scheduler.add_noise(model_input, noise, timesteps)
                noisy_model_input=noisy_model_input.to(accelerator.device)
                '''print('unet device',unet.device)
                print('noisy_model_input.device',noisy_model_input.device)
                print('timesteps.device',timesteps.device)
                print('prompts.device',prompts.device)'''
                model_pred = unet(
                    noisy_model_input, timesteps, prompts, return_dict=False
                )[0]
                if use_prior_preservation:
                    # Chunk the noise and model_pred into two parts and compute the loss on each part separately.
                    model_pred, model_pred_prior = torch.chunk(model_pred, 2, dim=0)
                    noise, noise_prior = torch.chunk(noise, 2, dim=0)
                    # Compute prior loss
                    prior_loss = F.mse_loss(model_pred_prior.float(), noise_prior.float(), reduction="mean")
                loss = F.mse_loss(model_pred.float(), noise.float(), reduction="mean")
                if use_prior_preservation:
                    loss=loss+prior_loss_weight*prior_loss

                accelerator.backward(loss)
                if accelerator.sync_gradients:
                    params_to_clip = (
                        unet.parameters()
                    )
                    accelerator.clip_grad_norm_(params_to_clip, max_grad_norm)
                    if track_loss:
                        try:
                            accelerator.log({
                                "loss":loss
                            })
                        except:
                            accelerator.log({
                                "loss":loss.detach().cpu().numpy()
                            })
                optimizer.step()
                optimizer.zero_grad()
        if accelerator.sync_gradients:
            progress_bar.update(1)
        if log_images!=-1 and e%log_images==0:
            image=pipeline(entity_name, num_inference_steps=noise_scheduler.config.num_train_timesteps).images[0]
            try:
                accelerator.log({
                    "training_image":wandb.Image(image)
                })
            except:
                path="temp_training.png"
                image.save(path)
                try:
                    accelerator.log({
                        "training_image":wandb.Image(path)
                    })
                except:
                    pass
    end=time.time()
    print(f"trained unet! time elapsed: {end-start}")
    return pipeline


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')

from src.experiment_helpers.legacy_dreamsim import dreamsim

dream_model, dream_preprocess = dreamsim(pretrained=True,cache_dir="/scratch/jlb638/dreamsim",device="cpu")


from segment_anything import SamPredictor, sam_model_registry
from PIL import Image
import cv2
import numpy as np

#NEVERMIND SAM KIND OF SUCKS BECAUSE IT SEGMENTS THINGS BUT WE DONT KNOW WHAT EACH THING IS

pretrained_path="/scratch/jlb638/segment_checkpoints/sam_vit_h_4b8939.pth"

def sam_mask(image:Image.Image, sam_predictor:SamPredictor,prompt:str)->Image.Image:
    cv_img=np.array(image)
    sam_predictor.set_image(cv_img)
    masks, _, _ = sam_predictor.predict(prompt)



from trl import DDPOConfig, DDPOTrainer, DefaultDDPOStableDiffusionPipeline,DDPOPipelineOutput
from peft import LoraConfig, get_peft_model
import torch
from typing import Any, Callable, Dict, List, Optional, Union 
from PIL import Image
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion import rescale_noise_cfg,StableDiffusionPipeline
from trl.models.modeling_sd_base import scheduler_step


class BetterDefaultDDPOStableDiffusionPipeline(DefaultDDPOStableDiffusionPipeline):
    def __init__(self,train_text_encoder:bool,
                 train_text_encoder_embeddings:bool,
                 train_unet:bool,
                  use_lora_text_encoder:bool, *args,**kwargs):
        super().__init__(*args, **kwargs)
        self.sd_pipeline("some things dont initialize correctly if you dont have this dumb step",num_inference_steps=1)
        self.sd_pipeline.safety_checker=None
        self.sd_pipeline.vae.requires_grad_(False)
        if train_text_encoder and train_text_encoder_embeddings:
            raise Exception("train text encoder OR embedding!!!")
        elif train_text_encoder:
            if use_lora_text_encoder:
                self.sd_pipeline.text_encoder.requires_grad_(False)
                text_encoder_target_modules=["q_proj", "v_proj"]
                text_encoder_config=LoraConfig(
                    r=8,
                    lora_alpha=32,
                    target_modules=text_encoder_target_modules,
                    lora_dropout=0.0
                )
                self.sd_pipeline.text_encoder=get_peft_model(self.sd_pipeline.text_encoder,text_encoder_config)
                self.sd_pipeline.text_encoder.print_trainable_parameters()
            else:
                self.sd_pipeline.text_encoder.requires_grad_(True)
        elif train_text_encoder_embeddings:
            self.sd_pipeline.text_encoder.requires_grad_(False)
            self.sd_pipeline.text_encoder.get_input_embeddings().requires_grad_(True)
        for param in self.sd_pipeline.text_encoder.parameters():
            if param.requires_grad:
                param.data = param.to(torch.float32)
        if train_unet:
            if self.use_lora:
                self.sd_pipeline.unet.requires_grad_(False)
                lora_config = LoraConfig(
                    r=8,
                    lora_alpha=32,
                    init_lora_weights="gaussian",
                    target_modules=["to_k", "to_q", "to_v", "to_out.0"],
                )
                self.sd_pipeline.unet=get_peft_model(self.sd_pipeline.unet,lora_config)
                self.sd_pipeline.unet.print_trainable_parameters()
                
            else:
                self.sd_pipeline.unet.requires_grad_(True)
            # To avoid accelerate unscaling problems in FP16.
            for param in self.sd_pipeline.unet.parameters():
                # only upcast trainable parameters (LoRA) into fp32
                if param.requires_grad:
                    param.data = param.to(torch.float32)

    def get_trainable_layers(self):
        return [p for p in self.sd_pipeline.unet.parameters() if p.requires_grad]+[p for p in self.sd_pipeline.text_encoder.parameters() if p.requires_grad]


import torch
from torch import nn

import inspect
import math
from typing import Callable, List, Optional, Tuple, Union,Dict
import torch.nn.functional as F



from diffusers.image_processor import IPAdapterMaskProcessor
from diffusers.utils import deprecate, logging
from diffusers.utils.torch_utils import is_torch_version, maybe_allow_in_graph
from diffusers.models.attention_processor import Attention

from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
import torch.nn.functional as F
from huggingface_hub.utils import validate_hf_hub_args
from safetensors import safe_open

from diffusers.models.modeling_utils import _LOW_CPU_MEM_USAGE_DEFAULT, load_state_dict
from diffusers.utils import (
    USE_PEFT_BACKEND,
    _get_model_file,
    is_accelerate_available,
    is_torch_version,
    is_transformers_available,
    logging,
)

from diffusers.models.modeling_utils import load_model_dict_into_meta, load_state_dict
from diffusers.loaders.unet_loader_utils import _maybe_expand_lora_scales
from diffusers.loaders.ip_adapter import IPAdapterMixin
from diffusers.loaders.unet import UNet2DConditionLoadersMixin
from diffusers.models.embeddings import (
    ImageProjection,
    IPAdapterFaceIDImageProjection,
    IPAdapterFaceIDPlusImageProjection,
    IPAdapterFullImageProjection,
    IPAdapterPlusImageProjection,
    MultiIPAdapterImageProjection,
)

from contextlib import nullcontext


if is_transformers_available():
    from transformers import (
        CLIPImageProcessor,
        CLIPVisionModelWithProjection,
    )

    from diffusers.models.attention_processor import (
        AttnProcessor,
        AttnProcessor2_0,
        IPAdapterAttnProcessor,
        IPAdapterAttnProcessor2_0,
    )

logger = logging.get_logger(__name__)

class IPAdapterAttnProcessorValue(IPAdapterAttnProcessor2_0): #https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py
    def __call__(
        self,
        attn: Attention,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        temb: Optional[torch.Tensor] = None,
        scale: float = 1.0,
        ip_adapter_masks: Optional[torch.Tensor] = None,
    ):
        residual = hidden_states

        # separate ip_hidden_states from encoder_hidden_states
        if encoder_hidden_states is not None:
            if isinstance(encoder_hidden_states, tuple):
                encoder_hidden_states, ip_hidden_states = encoder_hidden_states
            else:
                deprecation_message = (
                    "You have passed a tensor as `encoder_hidden_states`. This is deprecated and will be removed in a future release."
                    " Please make sure to update your script to pass `encoder_hidden_states` as a tuple to suppress this warning."
                )
                deprecate("encoder_hidden_states not a tuple", "1.0.0", deprecation_message, standard_warn=False)
                end_pos = encoder_hidden_states.shape[1] - self.num_tokens[0]
                encoder_hidden_states, ip_hidden_states = (
                    encoder_hidden_states[:, :end_pos, :],
                    [encoder_hidden_states[:, end_pos:, :]],
                )

        if attn.spatial_norm is not None:
            hidden_states = attn.spatial_norm(hidden_states, temb)

        input_ndim = hidden_states.ndim

        if input_ndim == 4:
            batch_size, channel, height, width = hidden_states.shape
            hidden_states = hidden_states.view(batch_size, channel, height * width).transpose(1, 2)

        batch_size, sequence_length, _ = (
            hidden_states.shape if encoder_hidden_states is None else encoder_hidden_states.shape
        )

        if attention_mask is not None:
            attention_mask = attn.prepare_attention_mask(attention_mask, sequence_length, batch_size)
            # scaled_dot_product_attention expects attention_mask shape to be
            # (batch, heads, source_length, target_length)
            attention_mask = attention_mask.view(batch_size, attn.heads, -1, attention_mask.shape[-1])

        if attn.group_norm is not None:
            hidden_states = attn.group_norm(hidden_states.transpose(1, 2)).transpose(1, 2)

        query = attn.to_q(hidden_states)

        if encoder_hidden_states is None:
            encoder_hidden_states = hidden_states
        elif attn.norm_cross:
            encoder_hidden_states = attn.norm_encoder_hidden_states(encoder_hidden_states)

        key = attn.to_k(encoder_hidden_states)
        value = attn.to_v(encoder_hidden_states)

        inner_dim = key.shape[-1]
        head_dim = inner_dim // attn.heads

        query = query.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        key = key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
        value = value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        # the output of sdp = (batch, num_heads, seq_len, head_dim)
        # TODO: add support for attn.scale when we move to Torch 2.1
        hidden_states = F.scaled_dot_product_attention(
            query, key, value, attn_mask=attention_mask, dropout_p=0.0, is_causal=False
        )

        hidden_states = hidden_states.transpose(1, 2).reshape(batch_size, -1, attn.heads * head_dim)
        hidden_states = hidden_states.to(query.dtype)

        if ip_adapter_masks is not None:
            if not isinstance(ip_adapter_masks, List):
                # for backward compatibility, we accept `ip_adapter_mask` as a tensor of shape [num_ip_adapter, 1, height, width]
                ip_adapter_masks = list(ip_adapter_masks.unsqueeze(1))
            if not (len(ip_adapter_masks) == len(self.scale) == len(ip_hidden_states)):
                raise ValueError(
                    f"Length of ip_adapter_masks array ({len(ip_adapter_masks)}) must match "
                    f"length of self.scale array ({len(self.scale)}) and number of ip_hidden_states "
                    f"({len(ip_hidden_states)})"
                )
            else:
                for index, (mask, scale, ip_state) in enumerate(zip(ip_adapter_masks, self.scale, ip_hidden_states)):
                    if not isinstance(mask, torch.Tensor) or mask.ndim != 4:
                        raise ValueError(
                            "Each element of the ip_adapter_masks array should be a tensor with shape "
                            "[1, num_images_for_ip_adapter, height, width]."
                            " Please use `IPAdapterMaskProcessor` to preprocess your mask"
                        )
                    if mask.shape[1] != ip_state.shape[1]:
                        raise ValueError(
                            f"Number of masks ({mask.shape[1]}) does not match "
                            f"number of ip images ({ip_state.shape[1]}) at index {index}"
                        )
                    if isinstance(scale, list) and not len(scale) == mask.shape[1]:
                        raise ValueError(
                            f"Number of masks ({mask.shape[1]}) does not match "
                            f"number of scales ({len(scale)}) at index {index}"
                        )
        else:
            ip_adapter_masks = [None] * len(self.scale)

        # for ip-adapter
        for current_ip_hidden_states, scale,  to_v_ip, mask in zip(
            ip_hidden_states, self.scale, self.to_v_ip, ip_adapter_masks
        ):
            skip = False
            if isinstance(scale, list):
                if all(s == 0 for s in scale):
                    skip = True
            elif scale == 0:
                skip = True
            if not skip:
                if mask is not None:
                    if not isinstance(scale, list):
                        scale = [scale] * mask.shape[1]

                    current_num_images = mask.shape[1]
                    for i in range(current_num_images):
                        #ip_key = to_k_ip(current_ip_hidden_states[:, i, :, :])
                        ip_value = to_v_ip(current_ip_hidden_states[:, i, :, :])

                        #ip_key = ip_key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
                        ip_value = ip_value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

                        # the output of sdp = (batch, num_heads, seq_len, head_dim)
                        # TODO: add support for attn.scale when we move to Torch 2.1
                        _current_ip_hidden_states = F.scaled_dot_product_attention(
                            query, key, ip_value, attn_mask=None, dropout_p=0.0, is_causal=False
                        )

                        _current_ip_hidden_states = _current_ip_hidden_states.transpose(1, 2).reshape(
                            batch_size, -1, attn.heads * head_dim
                        )
                        _current_ip_hidden_states = _current_ip_hidden_states.to(query.dtype)

                        mask_downsample = IPAdapterMaskProcessor.downsample(
                            mask[:, i, :, :],
                            batch_size,
                            _current_ip_hidden_states.shape[1],
                            _current_ip_hidden_states.shape[2],
                        )

                        mask_downsample = mask_downsample.to(dtype=query.dtype, device=query.device)
                        hidden_states = hidden_states + scale[i] * (_current_ip_hidden_states * mask_downsample)
                else:
                    #ip_key = to_k_ip(current_ip_hidden_states)
                    ip_value = to_v_ip(current_ip_hidden_states)

                    #ip_key = ip_key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
                    ip_value = ip_value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

                    # the output of sdp = (batch, num_heads, seq_len, head_dim)
                    # TODO: add support for attn.scale when we move to Torch 2.1

                    padding_size=(ip_value.size()[0], ip_value.size()[1],77-ip_value.size()[2],ip_value.size()[3])
                    padding_tensor=torch.ones(padding_size).to(ip_value.device)
                    ip_value=torch.cat((ip_value,padding_tensor),dim=2)

                    current_ip_hidden_states = F.scaled_dot_product_attention(
                        query, key, ip_value, attn_mask=None, dropout_p=0.0, is_causal=False
                    )

                    current_ip_hidden_states = current_ip_hidden_states.transpose(1, 2).reshape(
                        batch_size, -1, attn.heads * head_dim
                    )
                    current_ip_hidden_states = current_ip_hidden_states.to(query.dtype)

                    hidden_states = hidden_states + scale * current_ip_hidden_states

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if input_ndim == 4:
            hidden_states = hidden_states.transpose(-1, -2).reshape(batch_size, channel, height, width)

        if attn.residual_connection:
            hidden_states = hidden_states + residual

        hidden_states = hidden_states / attn.rescale_output_factor

        return hidden_states
    

class IPAdapterAttnProcessorKey(IPAdapterAttnProcessor2_0): #https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py
    def __call__(
        self,
        attn: Attention,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        temb: Optional[torch.Tensor] = None,
        scale: float = 1.0,
        ip_adapter_masks: Optional[torch.Tensor] = None,
    ):
        residual = hidden_states

        # separate ip_hidden_states from encoder_hidden_states
        if encoder_hidden_states is not None:
            if isinstance(encoder_hidden_states, tuple):
                encoder_hidden_states, ip_hidden_states = encoder_hidden_states
            else:
                deprecation_message = (
                    "You have passed a tensor as `encoder_hidden_states`. This is deprecated and will be removed in a future release."
                    " Please make sure to update your script to pass `encoder_hidden_states` as a tuple to suppress this warning."
                )
                deprecate("encoder_hidden_states not a tuple", "1.0.0", deprecation_message, standard_warn=False)
                end_pos = encoder_hidden_states.shape[1] - self.num_tokens[0]
                encoder_hidden_states, ip_hidden_states = (
                    encoder_hidden_states[:, :end_pos, :],
                    [encoder_hidden_states[:, end_pos:, :]],
                )

        if attn.spatial_norm is not None:
            hidden_states = attn.spatial_norm(hidden_states, temb)

        input_ndim = hidden_states.ndim

        if input_ndim == 4:
            batch_size, channel, height, width = hidden_states.shape
            hidden_states = hidden_states.view(batch_size, channel, height * width).transpose(1, 2)

        batch_size, sequence_length, _ = (
            hidden_states.shape if encoder_hidden_states is None else encoder_hidden_states.shape
        )

        if attention_mask is not None:
            attention_mask = attn.prepare_attention_mask(attention_mask, sequence_length, batch_size)
            # scaled_dot_product_attention expects attention_mask shape to be
            # (batch, heads, source_length, target_length)
            attention_mask = attention_mask.view(batch_size, attn.heads, -1, attention_mask.shape[-1])

        if attn.group_norm is not None:
            hidden_states = attn.group_norm(hidden_states.transpose(1, 2)).transpose(1, 2)

        query = attn.to_q(hidden_states)

        if encoder_hidden_states is None:
            encoder_hidden_states = hidden_states
        elif attn.norm_cross:
            encoder_hidden_states = attn.norm_encoder_hidden_states(encoder_hidden_states)

        key = attn.to_k(encoder_hidden_states)
        value = attn.to_v(encoder_hidden_states)

        inner_dim = key.shape[-1]
        head_dim = inner_dim // attn.heads

        query = query.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        key = key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
        value = value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        # the output of sdp = (batch, num_heads, seq_len, head_dim)
        # TODO: add support for attn.scale when we move to Torch 2.1
        hidden_states = F.scaled_dot_product_attention(
            query, key, value, attn_mask=attention_mask, dropout_p=0.0, is_causal=False
        )

        hidden_states = hidden_states.transpose(1, 2).reshape(batch_size, -1, attn.heads * head_dim)
        hidden_states = hidden_states.to(query.dtype)

        if ip_adapter_masks is not None:
            if not isinstance(ip_adapter_masks, List):
                # for backward compatibility, we accept `ip_adapter_mask` as a tensor of shape [num_ip_adapter, 1, height, width]
                ip_adapter_masks = list(ip_adapter_masks.unsqueeze(1))
            if not (len(ip_adapter_masks) == len(self.scale) == len(ip_hidden_states)):
                raise ValueError(
                    f"Length of ip_adapter_masks array ({len(ip_adapter_masks)}) must match "
                    f"length of self.scale array ({len(self.scale)}) and number of ip_hidden_states "
                    f"({len(ip_hidden_states)})"
                )
            else:
                for index, (mask, scale, ip_state) in enumerate(zip(ip_adapter_masks, self.scale, ip_hidden_states)):
                    if not isinstance(mask, torch.Tensor) or mask.ndim != 4:
                        raise ValueError(
                            "Each element of the ip_adapter_masks array should be a tensor with shape "
                            "[1, num_images_for_ip_adapter, height, width]."
                            " Please use `IPAdapterMaskProcessor` to preprocess your mask"
                        )
                    if mask.shape[1] != ip_state.shape[1]:
                        raise ValueError(
                            f"Number of masks ({mask.shape[1]}) does not match "
                            f"number of ip images ({ip_state.shape[1]}) at index {index}"
                        )
                    if isinstance(scale, list) and not len(scale) == mask.shape[1]:
                        raise ValueError(
                            f"Number of masks ({mask.shape[1]}) does not match "
                            f"number of scales ({len(scale)}) at index {index}"
                        )
        else:
            ip_adapter_masks = [None] * len(self.scale)

        # for ip-adapter
        for current_ip_hidden_states, scale,  to_k_ip, to_v_ip, mask in zip(
            ip_hidden_states, self.scale, self.to_k_ip, self.to_v_ip, ip_adapter_masks
        ):
            skip = False
            if isinstance(scale, list):
                if all(s == 0 for s in scale):
                    skip = True
            elif scale == 0:
                skip = True
            if not skip:
                if mask is not None:
                    if not isinstance(scale, list):
                        scale = [scale] * mask.shape[1]

                    current_num_images = mask.shape[1]
                    for i in range(current_num_images):
                        ip_key = to_k_ip(current_ip_hidden_states[:, i, :, :])
                        #ip_value = to_v_ip(current_ip_hidden_states[:, i, :, :])

                        ip_key = ip_key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
                        #ip_value = ip_value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

                        # the output of sdp = (batch, num_heads, seq_len, head_dim)
                        # TODO: add support for attn.scale when we move to Torch 2.1
                        _current_ip_hidden_states = F.scaled_dot_product_attention(
                            query, ip_key, value, attn_mask=None, dropout_p=0.0, is_causal=False
                        )

                        _current_ip_hidden_states = _current_ip_hidden_states.transpose(1, 2).reshape(
                            batch_size, -1, attn.heads * head_dim
                        )
                        _current_ip_hidden_states = _current_ip_hidden_states.to(query.dtype)

                        mask_downsample = IPAdapterMaskProcessor.downsample(
                            mask[:, i, :, :],
                            batch_size,
                            _current_ip_hidden_states.shape[1],
                            _current_ip_hidden_states.shape[2],
                        )

                        mask_downsample = mask_downsample.to(dtype=query.dtype, device=query.device)
                        hidden_states = hidden_states + scale[i] * (_current_ip_hidden_states * mask_downsample)
                else:
                    ip_key = to_k_ip(current_ip_hidden_states)
                    
                    #ip_value = to_v_ip(current_ip_hidden_states)

                    ip_key = ip_key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
                    #ip_value = ip_value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

                    #print("value size",value.size())
                    #print('key.size()',key.size())
                    #print('ip_value.size()',ip_value.size())
                    #print('ip_key.size()',ip_key.size())
                    padding_size=(ip_key.size()[0], ip_key.size()[1],77-ip_key.size()[2],ip_key.size()[3])
                    padding_tensor=torch.ones(padding_size).to(ip_key.device)
                    ip_key=torch.cat((ip_key,padding_tensor),dim=2)
                    # the output of sdp = (batch, num_heads, seq_len, head_dim)
                    # TODO: add support for attn.scale when we move to Torch 2.1
                    current_ip_hidden_states = F.scaled_dot_product_attention(
                        query, ip_key, value, attn_mask=None, dropout_p=0.0, is_causal=False
                    )

                    current_ip_hidden_states = current_ip_hidden_states.transpose(1, 2).reshape(
                        batch_size, -1, attn.heads * head_dim
                    )
                    current_ip_hidden_states = current_ip_hidden_states.to(query.dtype)

                    hidden_states = hidden_states + scale * current_ip_hidden_states

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if input_ndim == 4:
            hidden_states = hidden_states.transpose(-1, -2).reshape(batch_size, channel, height, width)

        if attn.residual_connection:
            hidden_states = hidden_states + residual

        hidden_states = hidden_states / attn.rescale_output_factor

        return hidden_states
    

def load_ip_adapter_single( #https://github.com/huggingface/diffusers/blob/main/src/diffusers/loaders/ip_adapter.py#L55
        self:IPAdapterMixin,
        pretrained_model_name_or_path_or_dict: Union[str, List[str], Dict[str, torch.Tensor]],
        subfolder: Union[str, List[str]],
        weight_name: Union[str, List[str]],
        variant: str,
        image_encoder_folder: Optional[str] = "image_encoder",
        **kwargs,):
    """
    Parameters:
        pretrained_model_name_or_path_or_dict (`str` or `List[str]` or `os.PathLike` or `List[os.PathLike]` or `dict` or `List[dict]`):
            Can be either:

                - A string, the *model id* (for example `google/ddpm-celebahq-256`) of a pretrained model hosted on
                    the Hub.
                - A path to a *directory* (for example `./my_model_directory`) containing the model weights saved
                    with [`ModelMixin.save_pretrained`].
                - A [torch state
                    dict](https://pytorch.org/tutorials/beginner/saving_loading_models.html#what-is-a-state-dict).
        subfolder (`str` or `List[str]`):
            The subfolder location of a model file within a larger model repository on the Hub or locally. If a
            list is passed, it should have the same length as `weight_name`.
        weight_name (`str` or `List[str]`):
            The name of the weight file to load. If a list is passed, it should have the same length as
            `weight_name`.
        variant (str):
            one of "key" or "value"
        image_encoder_folder (`str`, *optional*, defaults to `image_encoder`):
            The subfolder location of the image encoder within a larger model repository on the Hub or locally.
            Pass `None` to not load the image encoder. If the image encoder is located in a folder inside
            `subfolder`, you only need to pass the name of the folder that contains image encoder weights, e.g.
            `image_encoder_folder="image_encoder"`. If the image encoder is located in a folder other than
            `subfolder`, you should pass the path to the folder that contains image encoder weights, for example,
            `image_encoder_folder="different_subfolder/image_encoder"`.
        cache_dir (`Union[str, os.PathLike]`, *optional*):
            Path to a directory where a downloaded pretrained model configuration is cached if the standard cache
            is not used.
        force_download (`bool`, *optional*, defaults to `False`):
            Whether or not to force the (re-)download of the model weights and configuration files, overriding the
            cached versions if they exist.

        proxies (`Dict[str, str]`, *optional*):
            A dictionary of proxy servers to use by protocol or endpoint, for example, `{'http': 'foo.bar:3128',
            'http://hostname': 'foo.bar:4012'}`. The proxies are used on each request.
        local_files_only (`bool`, *optional*, defaults to `False`):
            Whether to only load local model weights and configuration files or not. If set to `True`, the model
            won't be downloaded from the Hub.
        token (`str` or *bool*, *optional*):
            The token to use as HTTP bearer authorization for remote files. If `True`, the token generated from
            `diffusers-cli login` (stored in `~/.huggingface`) is used.
        revision (`str`, *optional*, defaults to `"main"`):
            The specific model version to use. It can be a branch name, a tag name, a commit id, or any identifier
            allowed by Git.
        low_cpu_mem_usage (`bool`, *optional*, defaults to `True` if torch version >= 1.9.0 else `False`):
            Speed up model loading only loading the pretrained weights and not initializing the weights. This also
            tries to not use more than 1x model size in CPU memory (including peak memory) while loading the model.
            Only supported for PyTorch >= 1.9.0. If you are using an older version of PyTorch, setting this
            argument to `True` will raise an error.
    """

    # handle the list inputs for multiple IP Adapters
    if not isinstance(weight_name, list):
        weight_name = [weight_name]

    if not isinstance(pretrained_model_name_or_path_or_dict, list):
        pretrained_model_name_or_path_or_dict = [pretrained_model_name_or_path_or_dict]
    if len(pretrained_model_name_or_path_or_dict) == 1:
        pretrained_model_name_or_path_or_dict = pretrained_model_name_or_path_or_dict * len(weight_name)

    if not isinstance(subfolder, list):
        subfolder = [subfolder]
    if len(subfolder) == 1:
        subfolder = subfolder * len(weight_name)

    if len(weight_name) != len(pretrained_model_name_or_path_or_dict):
        raise ValueError("`weight_name` and `pretrained_model_name_or_path_or_dict` must have the same length.")

    if len(weight_name) != len(subfolder):
        raise ValueError("`weight_name` and `subfolder` must have the same length.")

    # Load the main state dict first.
    cache_dir = kwargs.pop("cache_dir", None)
    force_download = kwargs.pop("force_download", False)
    proxies = kwargs.pop("proxies", None)
    local_files_only = kwargs.pop("local_files_only", None)
    token = kwargs.pop("token", None)
    revision = kwargs.pop("revision", None)
    low_cpu_mem_usage = kwargs.pop("low_cpu_mem_usage", _LOW_CPU_MEM_USAGE_DEFAULT)

    if low_cpu_mem_usage and not is_accelerate_available():
        low_cpu_mem_usage = False
        logger.warning(
            "Cannot initialize model with low cpu memory usage because `accelerate` was not found in the"
            " environment. Defaulting to `low_cpu_mem_usage=False`. It is strongly recommended to install"
            " `accelerate` for faster and less memory-intense model loading. You can do so with: \n```\npip"
            " install accelerate\n```\n."
        )

    if low_cpu_mem_usage is True and not is_torch_version(">=", "1.9.0"):
        raise NotImplementedError(
            "Low memory initialization requires torch >= 1.9.0. Please either update your PyTorch version or set"
            " `low_cpu_mem_usage=False`."
        )

    user_agent = {
        "file_type": "attn_procs_weights",
        "framework": "pytorch",
    }
    state_dicts = []
    for pretrained_model_name_or_path_or_dict, weight_name, subfolder in zip(
        pretrained_model_name_or_path_or_dict, weight_name, subfolder
    ):
        if not isinstance(pretrained_model_name_or_path_or_dict, dict):
            model_file = _get_model_file(
                pretrained_model_name_or_path_or_dict,
                weights_name=weight_name,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                local_files_only=local_files_only,
                token=token,
                revision=revision,
                subfolder=subfolder,
                user_agent=user_agent,
            )
            if weight_name.endswith(".safetensors"):
                state_dict = {"image_proj": {}, "ip_adapter": {}}
                with safe_open(model_file, framework="pt", device="cpu") as f:
                    for key in f.keys():
                        if key.startswith("image_proj."):
                            state_dict["image_proj"][key.replace("image_proj.", "")] = f.get_tensor(key)
                        elif key.startswith("ip_adapter."):
                            state_dict["ip_adapter"][key.replace("ip_adapter.", "")] = f.get_tensor(key)
            else:
                state_dict = load_state_dict(model_file)
        else:
            state_dict = pretrained_model_name_or_path_or_dict

        keys = list(state_dict.keys())
        if keys != ["image_proj", "ip_adapter"]:
            raise ValueError("Required keys are (`image_proj` and `ip_adapter`) missing from the state dict.")

        state_dicts.append(state_dict)

        # load CLIP image encoder here if it has not been registered to the pipeline yet
        if hasattr(self, "image_encoder") and getattr(self, "image_encoder", None) is None:
            if image_encoder_folder is not None:
                if not isinstance(pretrained_model_name_or_path_or_dict, dict):
                    logger.info(f"loading image_encoder from {pretrained_model_name_or_path_or_dict}")
                    if image_encoder_folder.count("/") == 0:
                        image_encoder_subfolder = Path(subfolder, image_encoder_folder).as_posix()
                    else:
                        image_encoder_subfolder = Path(image_encoder_folder).as_posix()

                    image_encoder = CLIPVisionModelWithProjection.from_pretrained(
                        pretrained_model_name_or_path_or_dict,
                        subfolder=image_encoder_subfolder,
                        low_cpu_mem_usage=low_cpu_mem_usage,
                        cache_dir=cache_dir,
                        local_files_only=local_files_only,
                    ).to(self.device, dtype=self.dtype)
                    self.register_modules(image_encoder=image_encoder)
                else:
                    raise ValueError(
                        "`image_encoder` cannot be loaded because `pretrained_model_name_or_path_or_dict` is a state dict."
                    )
            else:
                logger.warning(
                    "image_encoder is not loaded since `image_encoder_folder=None` passed. You will not be able to use `ip_adapter_image` when calling the pipeline with IP-Adapter."
                    "Use `ip_adapter_image_embeds` to pass pre-generated image embedding instead."
                )

        # create feature extractor if it has not been registered to the pipeline yet
        if hasattr(self, "feature_extractor") and getattr(self, "feature_extractor", None) is None:
            # FaceID IP adapters don't need the image encoder so it's not present, in this case we default to 224
            default_clip_size = 224
            clip_image_size = (
                self.image_encoder.config.image_size if self.image_encoder is not None else default_clip_size
            )
            feature_extractor = CLIPImageProcessor(size=clip_image_size, crop_size=clip_image_size)
            self.register_modules(feature_extractor=feature_extractor)

    # load ip-adapter into unet
    unet = getattr(self, self.unet_name) if not hasattr(self, "unet") else self.unet
    #unet._load_ip_adapter_weights(state_dicts, low_cpu_mem_usage=low_cpu_mem_usage) #this we're changing tho
    _load_ip_adapter_weights_single(unet,state_dicts,variant,low_cpu_mem_usage)

    extra_loras = unet._load_ip_adapter_loras(state_dicts)
    if extra_loras != {}:
        if not USE_PEFT_BACKEND:
            logger.warning("PEFT backend is required to load these weights.")
        else:
            # apply the IP Adapter Face ID LoRA weights
            peft_config = getattr(unet, "peft_config", {})
            for k, lora in extra_loras.items():
                if f"faceid_{k}" not in peft_config:
                    self.load_lora_weights(lora, adapter_name=f"faceid_{k}")
                    self.set_adapters([f"faceid_{k}"], adapter_weights=[1.0])



def _load_ip_adapter_weights_single(self:UNet2DConditionLoadersMixin, state_dicts,variant:str, low_cpu_mem_usage:bool=False): #https://github.com/huggingface/diffusers/blob/main/src/diffusers/loaders/unet.py#L823
    if not isinstance(state_dicts, list):
        state_dicts = [state_dicts]

    # Kolors Unet already has a `encoder_hid_proj`
    if (
        self.encoder_hid_proj is not None
        and self.config.encoder_hid_dim_type == "text_proj"
        and not hasattr(self, "text_encoder_hid_proj")
    ):
        self.text_encoder_hid_proj = self.encoder_hid_proj

    # Set encoder_hid_proj after loading ip_adapter weights,
    # because `IPAdapterPlusImageProjection` also has `attn_processors`.
    self.encoder_hid_proj = None

    #attn_procs = self._convert_ip_adapter_attn_to_diffusers(state_dicts, low_cpu_mem_usage=low_cpu_mem_usage)
    attn_procs=_convert_ip_adapter_attn_to_diffusers_single(self,state_dicts,variant,low_cpu_mem_usage)
    self.set_attn_processor(attn_procs)

    # convert IP-Adapter Image Projection layers to diffusers
    image_projection_layers = []
    for state_dict in state_dicts:
        image_projection_layer = self._convert_ip_adapter_image_proj_to_diffusers(
            state_dict["image_proj"], low_cpu_mem_usage=low_cpu_mem_usage
        )
        image_projection_layers.append(image_projection_layer)

    self.encoder_hid_proj = MultiIPAdapterImageProjection(image_projection_layers)
    self.config.encoder_hid_dim_type = "ip_image_proj"

    self.to(dtype=self.dtype, device=self.device)


#https://github.com/huggingface/diffusers/blob/main/src/diffusers/loaders/unet.py#L733

def _convert_ip_adapter_attn_to_diffusers_single(self:UNet2DConditionLoadersMixin, state_dicts, variant:str, low_cpu_mem_usage=False):

    if low_cpu_mem_usage:
        if is_accelerate_available():
            from accelerate import init_empty_weights

        else:
            low_cpu_mem_usage = False
            logger.warning(
                "Cannot initialize model with low cpu memory usage because `accelerate` was not found in the"
                " environment. Defaulting to `low_cpu_mem_usage=False`. It is strongly recommended to install"
                " `accelerate` for faster and less memory-intense model loading. You can do so with: \n```\npip"
                " install accelerate\n```\n."
            )

    if low_cpu_mem_usage is True and not is_torch_version(">=", "1.9.0"):
        raise NotImplementedError(
            "Low memory initialization requires torch >= 1.9.0. Please either update your PyTorch version or set"
            " `low_cpu_mem_usage=False`."
        )

    # set ip-adapter cross-attention processors & load state_dict
    attn_procs = {}
    key_id = 1
    init_context = init_empty_weights if low_cpu_mem_usage else nullcontext
    for name in self.attn_processors.keys():
        cross_attention_dim = None if name.endswith("attn1.processor") else self.config.cross_attention_dim
        if name.startswith("mid_block"):
            hidden_size = self.config.block_out_channels[-1]
        elif name.startswith("up_blocks"):
            block_id = int(name[len("up_blocks.")])
            hidden_size = list(reversed(self.config.block_out_channels))[block_id]
        elif name.startswith("down_blocks"):
            block_id = int(name[len("down_blocks.")])
            hidden_size = self.config.block_out_channels[block_id]

        if cross_attention_dim is None or "motion_modules" in name:
            attn_processor_class = self.attn_processors[name].__class__
            attn_procs[name] = attn_processor_class()
        else:
            if variant=="key":
                attn_processor_class=IPAdapterAttnProcessorKey
            elif variant=="value":
                attn_processor_class=IPAdapterAttnProcessorValue
            num_image_text_embeds = []
            for state_dict in state_dicts:
                if "proj.weight" in state_dict["image_proj"]:
                    # IP-Adapter
                    num_image_text_embeds += [4]
                elif "proj.3.weight" in state_dict["image_proj"]:
                    # IP-Adapter Full Face
                    num_image_text_embeds += [257]  # 256 CLIP tokens + 1 CLS token
                elif "perceiver_resampler.proj_in.weight" in state_dict["image_proj"]:
                    # IP-Adapter Face ID Plus
                    num_image_text_embeds += [4]
                elif "norm.weight" in state_dict["image_proj"]:
                    # IP-Adapter Face ID
                    num_image_text_embeds += [4]
                else:
                    # IP-Adapter Plus
                    num_image_text_embeds += [state_dict["image_proj"]["latents"].shape[1]]

            with init_context():
                attn_procs[name] = attn_processor_class(
                    hidden_size=hidden_size,
                    cross_attention_dim=cross_attention_dim,
                    scale=1.0,
                    num_tokens=num_image_text_embeds,
                )

            value_dict = {}
            for i, state_dict in enumerate(state_dicts):
                value_dict.update({f"to_k_ip.{i}.weight": state_dict["ip_adapter"][f"{key_id}.to_k_ip.weight"]})
                value_dict.update({f"to_v_ip.{i}.weight": state_dict["ip_adapter"][f"{key_id}.to_v_ip.weight"]})

            if not low_cpu_mem_usage:
                attn_procs[name].load_state_dict(value_dict)
            else:
                device = next(iter(value_dict.values())).device
                dtype = next(iter(value_dict.values())).dtype
                load_model_dict_into_meta(attn_procs[name], value_dict, device=device, dtype=dtype)

            key_id += 2

    return attn_procs


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.elastic_face_iresnet import face_mask
from facenet_pytorch import MTCNN

from PIL import Image

src=Image.open("ArcaneJinx.jpg")
mtcnn=MTCNN(device="cpu")
masked=face_mask(src,mtcnn,10)
masked.save("masked_jinx.jpg")


import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from diffusers import StableDiffusionPipeline
from PIL import Image
import torch
from accelerate import Accelerator

from src.experiment_helpers.ip_attention_value import load_ip_adapter_single


image =Image.open("ArcaneJinx.jpg")

accelerator=Accelerator(log_with="wandb",gradient_accumulation_steps=4)
accelerator.init_trackers(project_name="testing_bullshit")




for v in ["key","value","vanilla","ip"]:
    print(v)
    gen=torch.Generator().manual_seed(42)
    pipeline=StableDiffusionPipeline.from_pretrained("runwayml/stable-diffusion-v1-5")

    if v=="key" or v=="value":

        load_ip_adapter_single(pipeline,"h94/IP-Adapter", subfolder="models", weight_name="ip-adapter_sd15.bin",variant=v)
        pipeline.set_ip_adapter_scale(0.9)
    elif v=="ip":
        pipeline.load_ip_adapter("h94/IP-Adapter", subfolder="models", weight_name="ip-adapter_sd15.bin")
        pipeline.set_ip_adapter_scale(0.9)

    pipeline=pipeline.to(accelerator.device)

    if v=="vanilla":
        new_image=pipeline("going for a walk", num_inference_steps=30).images[0]
    else:

        new_image=pipeline("going for a walk", num_inference_steps=30, ip_adapter_image=image).images[0]

    new_image.save(f"ip_test_{v}.jpg")





prompt_list= [
    "a photo of a {}",
    "a rendering of a {}",
    "a cropped photo of the {}",
    "the photo of a {}",
    "a photo of a clean {}",
    "a photo of a dirty {}",
    "a dark photo of the {}",
    "a photo of my {}",
    "a photo of the cool {}",
    "a close-up photo of a {}",
    "a bright photo of the {}",
    "a cropped photo of a {}",
    "a photo of the {}",
    "a good photo of the {}",
    "a photo of one {}",
    "a close-up photo of the {}",
    "a rendition of the {}",
    "a photo of the clean {}",
    "a rendition of a {}",
    "a photo of a nice {}",
    "a good photo of a {}",
    "a photo of the nice {}",
    "a photo of the small {}",
    "a photo of the weird {}",
    "a photo of the large {}",
    "a photo of a cool {}",
    "a photo of a small {}",
]

validation_prompt_list=[
    "a photo of {}"
]

evaluation_prompt_list=[
    "a photo of  {} at the beach",
    "a photo of  {} in the jungle",
    "a photo of  {} in the snow",
    "a photo of  {} in the street",
    "a photo of  {} with a city in the background",
    "a photo of  {} with a mountain in the background",
    "a photo of  {} with the Eiffel Tower in the background",
    "a photo of  {} near the Statue of Liberty",
    "a photo of  {} near the Sydney Opera House",
    "a photo of  {} floating on top of water",
    "a photo of  {} eating a burger",
    "a photo of  {} drinking a beer",
    "a photo of  {} wearing a blue hat",
    "a photo of  {} wearing sunglasses",
    "a photo of  {} playing with a ball",
    "a photo of  {} as a police officer"
]

long_evaluation_prompt_list = [
    "A photo of {} in a bustling commercial kitchen, surrounded by stainless steel countertops and filled with the aroma of sizzling spices:",
    "A photo of {} in a dimly lit jazz club, nestled in the heart of the city's historic district, where the walls echo with the soulful notes of a saxophone:",
    "A photo of {} in a foggy, cobblestone alleyway, with old brick buildings looming on either side, creating an atmosphere of mystery and intrigue:",
    "A photo of {} in a hidden library, tucked away in the depths of an ancient castle, where shafts of sunlight filter through stained glass windows to illuminate dusty tomes of magic:",
    "A photo of {} in a state-of-the-art research laboratory, filled with rows of gleaming stainless steel equipment and humming with the energy of groundbreaking scientific discovery:",
    "A photo of {} in a sun-drenched botanical garden, bursting with vibrant blooms of every color imaginable, with butterflies fluttering lazily through the air:",
    "A photo of {} in a quaint, corner café, with cozy mismatched furniture and walls lined with shelves of well-loved books, where the scent of freshly brewed coffee fills the air:",
    "A photo of {} on a secluded tropical beach, with powdery white sand stretching for miles in either direction, and the sound of crashing waves in the background:",
    "A photo of {} in a sunlit artist's studio, nestled in the attic of a charming cottage, with large windows overlooking a picturesque countryside scene:",
    "A photo of {} in a packed sports stadium, with fans cheering from the stands and a giant digital scoreboard displaying the latest world record-breaking feat:",
    "A photo of {} in a bustling urban metropolis, with skyscrapers towering overhead and the streets alive with the hustle and bustle of city life:",
    "A photo of {} in a vibrant, bustling classroom, adorned with colorful educational posters and filled with the eager chatter of students engaged in lively discussion:",
    "A photo of {} in a vast, golden wheat field, stretching as far as the eye can see, with a brilliant blue sky overhead and the distant silhouette of a farmhouse on the horizon:",
    "A photo of {} in an opulent ballroom, with crystal chandeliers casting a soft glow over the polished marble floors and couples waltzing gracefully across the room:",
    "A photo of {} in the depths of outer space, surrounded by the swirling colors of distant galaxies and the glittering light of a billion stars:"
]